"""Vocabulary class containing all logic for creating, checking, and updating vocabularies."""
import logging
from collections import OrderedDict
from os.path import join
from typing import Any, Dict, List, Optional, Tuple

from hdx.utilities.downloader import Download

from hdx.api.configuration import Configuration
from hdx.data.hdxobject import HDXObject

logger = logging.getLogger(__name__)


class ChainRuleError(Exception):
    pass


class Vocabulary(HDXObject):
    """Vocabulary class containing all logic for creating, checking, and updating vocabularies.

    Args:
        initial_data (Optional[Dict]): Initial dataset metadata dictionary. Defaults to None.
        name (str): Name of vocabulary
        tags (Optional[List]): Initial list of tags. Defaults to None.
        configuration (Optional[Configuration]): HDX configuration. Defaults to global configuration.
    """

    _approved_vocabulary = None
    _tags_dict = None

    def __init__(
        self,
        initial_data: Optional[Dict] = None,
        name: str = None,
        tags: Optional[List] = None,
        configuration: Optional[Configuration] = None,
    ) -> None:
        if not initial_data:
            initial_data = dict()
        if name:
            initial_data["name"] = name
        super().__init__(initial_data, configuration=configuration)
        if tags:
            self.add_tags(tags)

    @staticmethod
    def actions() -> Dict[str, str]:
        """Dictionary of actions that can be performed on object

        Returns:
            Dict[str, str]: Dictionary of actions that can be performed on object
        """
        return {
            "show": "vocabulary_show",
            "update": "vocabulary_update",
            "create": "vocabulary_create",
            "delete": "vocabulary_delete",
            "list": "vocabulary_list",
            "autocomplete": "tag_autocomplete",
        }

    def update_from_yaml(
        self, path: str = join("config", "hdx_vocabulary_static.yml")
    ) -> None:
        """Update vocabulary metadata with static metadata from YAML file

        Args:
            path (Optional[str]): Path to YAML dataset metadata. Defaults to config/hdx_vocabulary_static.yml.

        Returns:
            None
        """
        super().update_from_yaml(path)

    def update_from_json(
        self, path: str = join("config", "hdx_vocabulary_static.json")
    ) -> None:
        """Update vocabulary metadata with static metadata from JSON file

        Args:
            path (Optional[str]): Path to JSON dataset metadata. Defaults to config/hdx_vocabulary_static.json.

        Returns:
            None
        """
        super().update_from_json(path)

    @classmethod
    def read_from_hdx(
        cls, identifier: str, configuration: Optional[Configuration] = None
    ) -> Optional["Vocabulary"]:
        """Reads the vocabulary given by identifier from HDX and returns Vocabulary object

        Args:
            identifier (str): Identifier of vocabulary
            configuration (Optional[Configuration]): HDX configuration. Defaults to global configuration.

        Returns:
            Optional[Vocabulary]: Vocabulary object if successful read, None if not
        """
        return cls._read_from_hdx_class(
            "vocabulary", identifier, configuration
        )

    @staticmethod
    def get_all_vocabularies(
        configuration: Optional[Configuration] = None,
    ) -> List["Vocabulary"]:
        """Get all vocabulary names in HDX

        Args:
            identifier (str): Identifier of resource
            configuration (Optional[Configuration]): HDX configuration. Defaults to global configuration.

        Returns:
            List[Vocabulary]: List of Vocabulary objects
        """

        vocabulary = Vocabulary(configuration=configuration)
        vocabulary[
            "id"
        ] = "all vocabulary names"  # only for error message if produced
        vocabularies = list()
        for vocabularydict in vocabulary._write_to_hdx("list", {}):
            vocabularies.append(
                Vocabulary(vocabularydict, configuration=configuration)
            )
        return vocabularies

    def check_required_fields(self, ignore_fields: List[str] = list()) -> None:
        """Check that metadata for vocabulary is complete. The parameter ignore_fields should
        be set if required to any fields that should be ignored for the particular operation.

        Args:
            ignore_fields (List[str]): Fields to ignore. Default is [].

        Returns:
            None
        """
        self._check_required_fields("vocabulary", ignore_fields)

    def update_in_hdx(self, **kwargs: Any) -> None:
        """Check if vocabulary exists in HDX and if so, update vocabulary

        Returns:
            None
        """
        self._update_in_hdx("vocabulary", "id", force_active=False, **kwargs)

    def create_in_hdx(self, **kwargs: Any) -> None:
        """Check if vocabulary exists in HDX and if so, update it, otherwise create vocabulary

        Returns:
            None
        """
        self._create_in_hdx(
            "vocabulary", "id", "name", force_active=False, **kwargs
        )

    def delete_from_hdx(self, empty: bool = True) -> None:
        """Deletes a vocabulary from HDX. First tags are removed then vocabulary is deleted.

        Args:
            empty (bool): Remove all tags and update before deleting. Defaults to True.

        Returns:
            None
        """
        if empty and len(self.data["tags"]) != 0:
            self.data["tags"] = list()
            self._update_in_hdx(
                "vocabulary", "id", force_active=False, ignore_field="tags"
            )
        self._delete_from_hdx("vocabulary", "id")

    def get_tags(self) -> List[str]:
        """Lists tags in vocabulary

        Returns:
            List[str]: List of tags in vocabulary
        """
        return self._get_tags()

    def add_tag(self, tag: str) -> bool:
        """Add a tag

        Args:
            tag (str): Tag to add

        Returns:
            bool: True if tag added or False if tag already present
        """
        return self._add_tag(tag)

    def add_tags(self, tags: List[str]) -> List[str]:
        """Add a list of tags

        Args:
            tags (List[str]): list of tags to add

        Returns:
            List[str]: Tags that were successfully added
        """
        return self._add_tags(tags)

    def remove_tag(self, tag: str) -> bool:
        """Remove a tag

        Args:
            tag (str): Tag to remove

        Returns:
            bool: True if tag removed or False if not
        """
        return self._remove_hdxobject(
            self.data.get("tags"), tag.lower(), matchon="name"
        )

    @classmethod
    def get_approved_vocabulary(
        cls, configuration: Optional[Configuration] = None
    ) -> "Vocabulary":
        """
        Get the HDX approved vocabulary

        Args:
            configuration (Optional[Configuration]): HDX configuration. Defaults to global configuration.

        Returns:
            Vocabulary: HDX Vocabulary object
        """
        if cls._approved_vocabulary is None:
            if configuration is None:
                configuration = Configuration.read()
            vocabulary_name = configuration["approved_tags_vocabulary"]
            cls._approved_vocabulary = Vocabulary.read_from_hdx(
                vocabulary_name, configuration=configuration
            )
        return cls._approved_vocabulary

    @classmethod
    def _read_approved_tags(
        cls,
        url: Optional[str] = None,
        configuration: Optional[Configuration] = None,
    ) -> List[str]:
        """
        Read approved tags

        Args:
            url (Optional[str]): Url to read approved tags from. Defaults to None (url in internal configuration).
            configuration (Optional[Configuration]): HDX configuration. Defaults to global configuration.

        Returns:
            List[str]: List of approved tags
        """
        with Download(full_agent=configuration.get_user_agent()) as downloader:
            if url is None:
                url = configuration["tags_list_url"]
            return list(
                OrderedDict.fromkeys(
                    downloader.download(url).text.replace('"', "").splitlines()
                )
            )

    @classmethod
    def create_approved_vocabulary(
        cls,
        url: Optional[str] = None,
        configuration: Optional[Configuration] = None,
    ) -> "Vocabulary":
        """
        Create the HDX approved vocabulary

        Args:
            url (Optional[str]): Tag to check. Defaults to None (url in internal configuration).
            configuration (Optional[Configuration]): HDX configuration. Defaults to global configuration.

        Returns:
            Vocabulary: HDX Vocabulary object
        """
        if configuration is None:
            configuration = Configuration.read()
        vocabulary_name = configuration["approved_tags_vocabulary"]
        tags = cls._read_approved_tags(url=url, configuration=configuration)
        cls._approved_vocabulary = Vocabulary(
            name=vocabulary_name, tags=tags, configuration=configuration
        )
        cls._approved_vocabulary.create_in_hdx()
        return cls._approved_vocabulary

    @classmethod
    def update_approved_vocabulary(
        cls,
        url: Optional[str] = None,
        configuration: Optional[Configuration] = None,
    ) -> "Vocabulary":
        """
        Update the HDX approved vocabulary

        Args:
            url (Optional[str]): Tag to check. Defaults to None (url in internal configuration).
            configuration (Optional[Configuration]): HDX configuration. Defaults to global configuration.

        Returns:
            Vocabulary: HDX Vocabulary object
        """
        if configuration is None:
            configuration = Configuration.read()
        vocabulary = cls.get_approved_vocabulary(configuration=configuration)
        vocabulary["tags"] = list()
        vocabulary.add_tags(
            cls._read_approved_tags(url=url, configuration=configuration)
        )
        vocabulary.update_in_hdx()
        return vocabulary

    @classmethod
    def delete_approved_vocabulary(
        cls, configuration: Optional[Configuration] = None
    ) -> None:
        """
        Delete the approved vocabulary

        Args:
            configuration (Optional[Configuration]): HDX configuration. Defaults to global configuration.

        Returns:
            None
        """
        if configuration is None:
            configuration = Configuration.read()
        vocabulary = cls.get_approved_vocabulary(configuration=configuration)
        vocabulary.delete_from_hdx()

    @classmethod
    def approved_tags(
        cls, configuration: Optional[Configuration] = None
    ) -> List[str]:
        """
        Return list of approved tags

        Args:
            configuration (Optional[Configuration]): HDX configuration. Defaults to global configuration.

        Returns:
            List[str]: List of approved tags
        """
        return [
            x["name"]
            for x in cls.get_approved_vocabulary(configuration=configuration)[
                "tags"
            ]
        ]

    @classmethod
    def is_approved(
        cls, tag: str, configuration: Optional[Configuration] = None
    ) -> bool:
        """
        Return if tag is an approved one or not

        Args:
            tag (str): Tag to check
            configuration (Optional[Configuration]): HDX configuration. Defaults to global configuration.

        Returns:
            bool: True if tag is approved, False if not
        """
        if tag.lower() in cls.approved_tags(configuration):
            return True
        return False

    @classmethod
    def read_tags_mappings(
        cls,
        configuration: Optional[Configuration] = None,
        url: Optional[str] = None,
        keycolumn: int = 1,
        failchained: bool = True,
    ) -> Dict:
        """
        Read tag mappings and setup tags cleanup dictionaries

        Args:
            configuration (Optional[Configuration]): HDX configuration. Defaults to global configuration.
            url (Optional[str]): Url of tags cleanup spreadsheet. Defaults to None (internal configuration parameter).
            keycolumn (int): Column number of tag column in spreadsheet. Defaults to 1.
            failchained (bool): Fail if chained rules found. Defaults to True.

        Returns:
            Dict: Returns Tags dictionary
        """
        if not cls._tags_dict:
            if configuration is None:
                configuration = Configuration.read()
            with Download(
                full_agent=configuration.get_user_agent()
            ) as downloader:
                if url is None:
                    url = configuration["tags_mapping_url"]
                cls._tags_dict = downloader.download_tabular_rows_as_dicts(
                    url, keycolumn=keycolumn
                )
                keys = cls._tags_dict.keys()
                chainerror = False
                for i, tag in enumerate(keys):
                    whattodo = cls._tags_dict[tag]
                    final_tags = whattodo["New Tag(s)"]
                    if final_tags is None:
                        continue
                    action = whattodo["Action to Take"]
                    for final_tag in final_tags.split(";"):
                        if final_tag in keys:
                            index = list(keys).index(final_tag)
                            if index != i:
                                whattodo2 = cls._tags_dict[final_tag]
                                action2 = whattodo2["Action to Take"]
                                if action2 != "ok" and action2 != "other":
                                    final_tags2 = whattodo2["New Tag(s)"]
                                    if final_tag not in final_tags2.split(";"):
                                        chainerror = True
                                        if failchained:
                                            logger.error(
                                                "Chained rules: %s (%s -> %s) | %s (%s -> %s)"
                                                % (
                                                    action,
                                                    tag,
                                                    final_tags,
                                                    action2,
                                                    final_tag,
                                                    final_tags2,
                                                )
                                            )

                if failchained and chainerror:
                    raise ChainRuleError("Chained rules for tags detected!")
        return cls._tags_dict

    @classmethod
    def set_tagsdict(cls, tags_dict: Dict) -> None:
        """
        Set tags dictionary

        Args:
            tags_dict (Dict): Tags dictionary

        Returns:
            None
        """
        cls._tags_dict = tags_dict

    @classmethod
    def get_mapped_tag(
        cls,
        tag: str,
        log_deleted: bool = True,
        configuration: Optional[Configuration] = None,
    ) -> Tuple[List[str], List[str]]:
        """Given a tag, return a list of tag(s) to which it maps and any deleted tags

        Args:
            tags (str): Tag to map
            log_deleted (bool): Whether to log informational messages about deleted tags. Defaults to True.
            configuration (Optional[Configuration]): HDX configuration. Defaults to global configuration.

        Returns:
            Tuple[List[str], List[str]]: Tuple containing list of mapped tag(s) and list of deleted tags
        """
        if configuration is None:
            configuration = Configuration.read()
        tag = tag.lower()
        tags_dict = cls.read_tags_mappings(configuration=configuration)
        tags = list()
        deleted_tags = list()
        if cls.is_approved(tag):
            tags.append(tag)
        elif tag not in tags_dict.keys():
            logger.error(
                f"Unapproved tag {tag} not in tag mapping! For a list of approved tags see: {configuration['tags_list_url']}"
            )
            deleted_tags.append(tag)
        else:
            whattodo = tags_dict[tag]
            action = whattodo["Action to Take"]
            if action == "ok":
                logger.error(
                    f"Tag {tag} is not in CKAN approved tags but is in tags mappings! For a list of approved tags see: {configuration['tags_list_url']}"
                )
            elif action == "delete":
                if log_deleted:
                    logger.info(
                        f"Tag {tag} is invalid and won't be added! For a list of approved tags see: {configuration['tags_list_url']}"
                    )
                deleted_tags.append(tag)
            elif action == "merge":
                final_tags = whattodo["New Tag(s)"].split(";")
                tags.extend(final_tags)
            else:
                logger.error(f"Invalid action {action}!")
        return tags, deleted_tags

    @classmethod
    def get_mapped_tags(
        cls,
        tags: List[str],
        log_deleted: bool = True,
        configuration: Optional[Configuration] = None,
    ) -> Tuple[List[str], List[str]]:
        """Given a list of tags, return a list of tags to which they map and any deleted tags

        Args:
            tags (List[str]): List of tags to map
            log_deleted (bool): Whether to log informational messages about deleted tags. Defaults to True.
            configuration (Optional[Configuration]): HDX configuration. Defaults to global configuration.

        Returns:
            Tuple[List[str], List[str]]: Tuple containing list of mapped tags and list of deleted tags
        """
        new_tags = list()
        deleted_tags = list()
        for tag in tags:
            mapped_tags, del_tags = cls.get_mapped_tag(
                tag, log_deleted=log_deleted, configuration=configuration
            )
            new_tags.extend(mapped_tags)
            deleted_tags.extend(del_tags)
        return list(OrderedDict.fromkeys(new_tags)), list(
            OrderedDict.fromkeys(deleted_tags)
        )

    @classmethod
    def add_mapped_tag(
        cls, hdxobject: HDXObject, tag: str, log_deleted: bool = True
    ) -> Tuple[List[str], List[str]]:
        """Add a tag to an HDX object that has tags

        Args:
            hdxobject (HDXObject): HDX object such as dataset
            tag (str): Tag to add
            log_deleted (bool): Whether to log informational messages about deleted tags. Defaults to True.

        Returns:
            Tuple[List[str], List[str]]: Tuple containing list of added tags and list of deleted tags and tags not added
        """
        return cls.add_mapped_tags(hdxobject, [tag], log_deleted=log_deleted)

    @classmethod
    def add_mapped_tags(
        cls,
        hdxobject: HDXObject,
        tags: List[str],
        log_deleted: bool = True,
    ) -> Tuple[List[str], List[str]]:
        """Add a list of tag to an HDX object that has tags

        Args:
            hdxobject (HDXObject): HDX object such as dataset
            tags (List[str]): List of tags to add
            log_deleted (bool): Whether to log informational messages about deleted tags. Defaults to True.

        Returns:
            Tuple[List[str], List[str]]: Tuple containing list of added tags and list of deleted tags and tags not added
        """
        new_tags, deleted_tags = cls.get_mapped_tags(
            tags,
            log_deleted=log_deleted,
            configuration=hdxobject.configuration,
        )
        added_tags = hdxobject._add_tags(
            new_tags,
            cls.get_approved_vocabulary(configuration=hdxobject.configuration)[
                "id"
            ],
        )
        unadded_tags = [x for x in new_tags if x not in added_tags]
        unadded_tags.extend(deleted_tags)
        return added_tags, unadded_tags

    @classmethod
    def clean_tags(
        cls, hdxobject: HDXObject, log_deleted: bool = True
    ) -> Tuple[List[str], List[str]]:
        """Clean tags in an HDX object according to tags cleanup spreadsheet

        Args:
            hdxobject (HDXObject): HDX object such as dataset
            log_deleted (bool): Whether to log informational messages about deleted tags. Defaults to True.

        Returns:
            Tuple[List[str], List[str]]: Tuple containing list of mapped tags and list of deleted tags and tags not added
        """
        tags = hdxobject._get_tags()
        hdxobject["tags"] = list()
        return cls.add_mapped_tags(hdxobject, tags, log_deleted=log_deleted)

    @classmethod
    def autocomplete(
        cls,
        name: str,
        limit: int = 20,
        configuration: Optional[Configuration] = None,
        **kwargs: Any,
    ) -> List:
        """Autocomplete a tag name and return matches

        Args:
            name (str): Name to autocomplete
            limit (int): Maximum number of matches to return
            configuration (Optional[Configuration]): HDX configuration. Defaults to global configuration.
            **kwargs:
            offset (int): The offset to start returning tags from.

        Returns:
            List: Autocomplete matches
        """
        return cls._autocomplete(name, limit, configuration, **kwargs)



"""Helper to the Dataset class for handling resources with filestores.
"""
from typing import TYPE_CHECKING, Any, Dict

from hdx.utilities.dictandlist import merge_two_dictionaries

if TYPE_CHECKING:
    from hdx.data.resource import Resource


class FilestoreHelper:
    temporary_url = "updated_by_file_upload_step"

    @staticmethod
    def resource_check_required_fields(
        resource: "Resource", check_upload: bool = False, **kwargs: Any
    ) -> None:
        """Helper method to get ignore_fields from kwargs if it exists and add package_id

        Args:
            resource (Resource): Resource to check
            check_upload (bool): Whether to check for file upload. Defaults to False.
            **kwargs: Keyword arguments

        Returns:
            None
        """
        if "ignore_check" in kwargs:  # allow ignoring of field checks
            return
        if (
            check_upload
            and resource.get_file_to_upload()
            and "url" in resource.data
        ):
            del resource.data["url"]
        ignore_fields = kwargs.get("ignore_fields", list())
        resource_ignore_fields = list()
        for ignore_field in ignore_fields:
            if ignore_field.startswith("resource:"):
                resource_ignore_field = ignore_field[9:].strip()
                resource_ignore_fields.append(resource_ignore_field)
        if "package_id" not in resource_ignore_fields:
            resource_ignore_fields.append("package_id")
        resource.check_required_fields(ignore_fields=resource_ignore_fields)

    @classmethod
    def check_filestore_resource(
        cls,
        resource: "Resource",
        filestore_resources: Dict[int, str],
        resource_index: int,
        **kwargs: Any,
    ) -> None:
        """Helper method to add new resource from dataset including filestore.

        Args:
            resource (Resource): Resource to check
            filestore_resources (Dict[int, str]): List of (index of resource, file to upload)
            resource_index (int): Index of resource

        Returns:
            None
        """
        cls.resource_check_required_fields(resource, **kwargs)
        file_to_upload = resource.get_file_to_upload()
        if file_to_upload:
            filestore_resources[resource_index] = file_to_upload
            resource["url"] = cls.temporary_url

    @classmethod
    def dataset_merge_filestore_resource(
        cls,
        resource: "Resource",
        updated_resource: "Resource",
        filestore_resources: Dict[int, str],
        resource_index: int,
        **kwargs: Any,
    ) -> None:
        """Helper method to merge updated resource from dataset into HDX resource read from HDX including filestore.

        Args:
            resource (Resource): Resource read from HDX
            updated_resource (Resource): Updated resource from dataset
            filestore_resources (Dict[int, str]): List of (index of resources, file to upload)
            resource_index (int): Index of resource

        Returns:
            None
        """
        file_to_upload = updated_resource.get_file_to_upload()
        if file_to_upload:
            resource.set_file_to_upload(file_to_upload)
            filestore_resources[resource_index] = file_to_upload
        merge_two_dictionaries(resource, updated_resource)
        cls.resource_check_required_fields(
            resource, check_upload=True, **kwargs
        )
        if resource.get_file_to_upload():
            resource["url"] = cls.temporary_url



"""Dataset Tests (noncore methods)"""
import copy
import datetime
import json
from os.path import join

import pytest
from hdx.location.country import Country
from hdx.utilities.compare import assert_files_same
from hdx.utilities.dateparse import parse_date_range
from hdx.utilities.downloader import Download
from hdx.utilities.path import temp_dir
from hdx.utilities.saver import save_text

from hdx.api.configuration import Configuration
from hdx.data.dataset import Dataset
from hdx.data.hdxobject import HDXError
from hdx.data.organization import Organization
from hdx.data.user import User
from hdx.data.vocabulary import Vocabulary

from . import (
    MockResponse,
    dataset_data,
    dataset_mockshow,
    organization_data,
    resources_data,
    resultgroups,
    resulttags,
    user_data,
)
from .test_organization import organization_mockshow
from .test_resource_view import (
    resource_view_list,
    resource_view_mockcreate,
    resource_view_mocklist,
    resource_view_mockshow,
)
from .test_showcase import showcase_resultdict
from .test_user import user_mockshow
from .test_vocabulary import vocabulary_mockshow


class TestDatasetNoncore:
    association = None
    url = "https://raw.githubusercontent.com/OCHA-DAP/hdx-python-api/master/tests/fixtures/test_data.csv"
    hxltags = {
        "EVENT_ID_CNTY": "#event+code",
        "EVENT_DATE": "#date+occurred",
        "YEAR": "#date+year",
        "EVENT_TYPE": "#event+type",
        "ACTOR1": "#group+name+first",
        "ASSOC_ACTOR_1": "#group+name+first+assoc",
        "ACTOR2": "#group+name+second",
        "ASSOC_ACTOR_2": "#group+name+second+assoc",
        "REGION": "#region+name",
        "COUNTRY": "#country+name",
        "ADMIN1": "#adm1+name",
        "ADMIN2": "#adm2+name",
        "ADMIN3": "#adm3+name",
        "LOCATION": "#loc+name",
        "LATITUDE": "#geo+lat",
        "LONGITUDE": "#geo+lon",
        "SOURCE": "#meta+source",
        "NOTES": "#description",
        "FATALITIES": "#affected+killed",
        "ISO3": "#country+code",
    }

    @pytest.fixture(scope="class")
    def static_resource_view_yaml(self):
        return join(
            "tests", "fixtures", "config", "hdx_resource_view_static.yml"
        )

    @pytest.fixture(scope="function")
    def vocabulary_read(self):
        class MockSession:
            @staticmethod
            def post(url, data, headers, files, allow_redirects, auth=None):
                datadict = json.loads(data.decode("utf-8"))
                return vocabulary_mockshow(url, datadict)

        Configuration.read().remoteckan().session = MockSession()

    @pytest.fixture(scope="function")
    def user_read(self):
        class MockSession:
            @staticmethod
            def post(url, data, headers, files, allow_redirects, auth=None):
                datadict = json.loads(data.decode("utf-8"))
                return user_mockshow(url, datadict)

        Configuration.read().remoteckan().session = MockSession()

    @pytest.fixture(scope="function")
    def organization_read(self):
        class MockSession:
            @staticmethod
            def post(url, data, headers, files, allow_redirects, auth=None):
                datadict = json.loads(data.decode("utf-8"))
                return organization_mockshow(url, datadict)

        Configuration.read().remoteckan().session = MockSession()

    @pytest.fixture(scope="function")
    def showcase_read(self):
        class MockSession:
            @staticmethod
            def post(url, data, headers, files, allow_redirects, auth=None):
                datadict = json.loads(data.decode("utf-8"))
                if "showcase_list" in url:
                    result = json.dumps([showcase_resultdict])
                    return MockResponse(
                        200,
                        '{"success": true, "result": %s, "help": "http://test-data.humdata.org/api/3/action/help_show?name=ckanext_package_showcase_list"}'
                        % result,
                    )
                if "association_delete" in url:
                    TestDatasetNoncore.association = "delete"
                    return MockResponse(
                        200,
                        '{"success": true, "result": null, "help": "http://test-data.humdata.org/api/3/action/help_show?name=ckanext_showcase_package_association_delete"}',
                    )
                elif "association_create" in url:
                    TestDatasetNoncore.association = "create"
                    result = json.dumps(datadict)
                    return MockResponse(
                        200,
                        '{"success": true, "result": %s, "help": "http://test-data.humdata.org/api/3/action/help_show?name=ckanext_showcase_package_association_create"}'
                        % result,
                    )
                return dataset_mockshow(url, datadict)

        Configuration.read().remoteckan().session = MockSession()

    @pytest.fixture(scope="function")
    def vocabulary_update(self):
        class MockSession:
            @staticmethod
            def post(url, data, headers, files, allow_redirects, auth=None):
                if isinstance(data, dict):
                    datadict = {
                        k.decode("utf8"): v.decode("utf8")
                        for k, v in data.items()
                    }
                else:
                    datadict = json.loads(data.decode("utf-8"))
                if "default" in url:
                    result = json.dumps(resource_view_list)
                    return MockResponse(
                        200,
                        '{"success": true, "result": %s, "help": "http://test-data.humdata.org/api/3/action/help_show?name=package_create_default_resource_views"}'
                        % result,
                    )
                if "resource_view" in url:
                    if "show" in url:
                        return resource_view_mockshow(url, datadict)
                    if "list" in url:
                        return resource_view_mocklist(url, datadict)
                    if "create" in url:
                        if datadict["title"] == "Quick Charts":
                            return resource_view_mockcreate(url, datadict)
                    return MockResponse(
                        404,
                        '{"success": false, "error": {"message": "TEST ERROR: Not create", "__type": "TEST ERROR: Not Create Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=resource_view_create"}',
                    )

        Configuration.read().remoteckan().session = MockSession()

    def test_get_hdx_url(
        self, configuration, hdx_config_yaml, project_config_yaml
    ):
        dataset = Dataset()
        assert dataset.get_hdx_url() is None
        datasetdata = copy.deepcopy(dataset_data)
        dataset = Dataset(datasetdata)
        assert (
            dataset.get_hdx_url()
            == "https://data.humdata.org/dataset/MyDataset1"
        )
        Configuration.delete()
        Configuration._create(
            hdx_site="feature",
            user_agent="test",
            hdx_config_yaml=hdx_config_yaml,
            project_config_yaml=project_config_yaml,
        )
        dataset = Dataset(datasetdata)
        assert (
            dataset.get_hdx_url()
            == "https://feature.data-humdata-org.ahconu.org/dataset/MyDataset1"
        )

    def test_get_set_date_of_dataset(self):
        dataset = Dataset({"dataset_date": "[2020-01-07T00:00:00 TO *]"})
        result = dataset.get_date_of_dataset(today=datetime.date(2020, 11, 17))
        assert result == {
            "startdate": datetime.datetime(2020, 1, 7, 0, 0),
            "enddate": datetime.datetime(2020, 11, 17, 0, 0),
            "startdate_str": "2020-01-07T00:00:00",
            "enddate_str": "2020-11-17T00:00:00",
            "ongoing": True,
        }
        dataset.set_date_of_dataset("2020-02-09")
        result = dataset.get_date_of_dataset("%d/%m/%Y")
        assert result == {
            "startdate": datetime.datetime(2020, 2, 9, 0, 0),
            "enddate": datetime.datetime(2020, 2, 9, 0, 0),
            "startdate_str": "09/02/2020",
            "enddate_str": "09/02/2020",
            "ongoing": False,
        }
        dataset.set_date_of_dataset("2020-02-09", "2020-10-20")
        result = dataset.get_date_of_dataset("%d/%m/%Y")
        assert result == {
            "startdate": datetime.datetime(2020, 2, 9, 0, 0),
            "enddate": datetime.datetime(2020, 10, 20, 0, 0),
            "startdate_str": "09/02/2020",
            "enddate_str": "20/10/2020",
            "ongoing": False,
        }
        dataset.set_date_of_dataset("2020-02-09", ongoing=True)
        result = dataset.get_date_of_dataset(
            "%d/%m/%Y", today=datetime.datetime(2020, 3, 9, 0, 0)
        )
        assert result == {
            "startdate": datetime.datetime(2020, 2, 9, 0, 0),
            "enddate": datetime.datetime(2020, 3, 9, 0, 0),
            "startdate_str": "09/02/2020",
            "enddate_str": "09/03/2020",
            "ongoing": True,
        }

    def test_set_dataset_year_range(self, configuration):
        dataset = Dataset()
        retval = dataset.set_dataset_year_range(2001, 2015)
        assert retval == [2001, 2015]
        retval = dataset.set_dataset_year_range("2010", "2017")
        assert retval == [2010, 2017]
        retval = dataset.set_dataset_year_range("2013")
        assert retval == [2013]
        retval = dataset.set_dataset_year_range({2005, 2002, 2003})
        assert retval == [2002, 2003, 2005]
        retval = dataset.set_dataset_year_range([2005, 2002, 2003])
        assert retval == [2002, 2003, 2005]
        retval = dataset.set_dataset_year_range((2005, 2002, 2003))
        assert retval == [2002, 2003, 2005]

    def test_is_set_subnational(self):
        datasetdata = copy.deepcopy(dataset_data)
        dataset = Dataset(datasetdata)
        assert dataset["subnational"] == "1"
        assert dataset.is_subnational() is True
        dataset.set_subnational(False)
        assert dataset["subnational"] == "0"
        assert dataset.is_subnational() is False
        dataset.set_subnational(True)
        assert dataset["subnational"] == "1"
        assert dataset.is_subnational() is True

    def test_get_add_location(self, locations):
        Country.countriesdata(use_live=False)
        datasetdata = copy.deepcopy(dataset_data)
        dataset = Dataset(datasetdata)
        assert dataset["groups"] == resultgroups
        assert dataset.get_location_names() == ["Algeria", "Zimbabwe"]
        dataset.add_country_location("sdn")
        expected = copy.deepcopy(resultgroups)
        expected.append({"name": "sdn"})
        assert dataset["groups"] == expected
        assert dataset.get_location_names() == ["Algeria", "Zimbabwe", "Sudan"]
        dataset.add_country_location("dza")
        assert dataset["groups"] == expected
        assert dataset.get_location_names() == ["Algeria", "Zimbabwe", "Sudan"]
        dataset.add_country_locations(["KEN", "Mozambique", "ken"])
        expected.extend([{"name": "ken"}, {"name": "moz"}])
        assert dataset["groups"] == expected
        assert dataset.get_location_names() == [
            "Algeria",
            "Zimbabwe",
            "Sudan",
            "Kenya",
            "Mozambique",
        ]
        assert dataset.get_location_iso3s() == [
            "dza",
            "zwe",
            "sdn",
            "ken",
            "moz",
        ]
        dataset.remove_location("sdn")
        assert dataset.get_location_names() == [
            "Algeria",
            "Zimbabwe",
            "Kenya",
            "Mozambique",
        ]
        with pytest.raises(HDXError):
            dataset.add_region_location("NOTEXIST")
        dataset.add_region_location("Africa")
        assert len(dataset["groups"]) == 60
        assert len(dataset.get_location_names()) == 60
        del dataset["groups"]
        assert dataset.get_location_names() == []
        assert dataset.get_location_iso3s() == []
        with pytest.raises(HDXError):
            dataset.add_country_location("abc")
        with pytest.raises(HDXError):
            dataset.add_country_location("lala")
        dataset.add_country_location("Ukrai", exact=False)
        assert dataset["groups"] == [{"name": "ukr"}]
        assert dataset.get_location_names() == ["Ukraine"]
        dataset.add_country_location("ukr")
        dataset.add_other_location("nepal-earthquake")
        assert dataset["groups"] == [
            {"name": "ukr"},
            {"name": "nepal-earthquake"},
        ]
        assert dataset.get_location_names() == ["Ukraine", "Nepal Earthquake"]
        del dataset["groups"]
        dataset.add_other_location("Nepal E", exact=False)
        assert dataset["groups"] == [{"name": "nepal-earthquake"}]
        dataset.add_other_location("Nepal Earthquake")
        assert dataset["groups"] == [{"name": "nepal-earthquake"}]
        with pytest.raises(HDXError):
            dataset.add_other_location("lala")
        with pytest.raises(HDXError):
            dataset.add_other_location("lala", alterror="nana")
        dataset["groups"] = [{"name": "ken"}, {"name": "MOZ"}, {"name": "dza"}]
        dataset.remove_location("moz")
        assert dataset["groups"] == [{"name": "ken"}, {"name": "dza"}]
        dataset.remove_location("KEN")
        assert dataset["groups"] == [{"name": "dza"}]

    def test_transform_update_frequency(self):
        assert len(Dataset.list_valid_update_frequencies()) == 32
        assert Dataset.transform_update_frequency("-2") == "As needed"
        assert Dataset.transform_update_frequency("-1") == "Never"
        assert Dataset.transform_update_frequency("0") == "Live"
        assert Dataset.transform_update_frequency("1") == "Every day"
        assert Dataset.transform_update_frequency("Adhoc") == "-2"
        assert Dataset.transform_update_frequency("As needed") == "-2"
        assert Dataset.transform_update_frequency("Never") == "-1"
        assert Dataset.transform_update_frequency("Live") == "0"
        assert Dataset.transform_update_frequency("Every day") == "1"
        assert Dataset.transform_update_frequency("EVERY WEEK") == "7"
        assert Dataset.transform_update_frequency("every month") == "30"
        assert Dataset.transform_update_frequency("LALA") is None
        assert Dataset.transform_update_frequency(-2) == "As needed"
        assert Dataset.transform_update_frequency(7) == "Every week"
        assert Dataset.transform_update_frequency("") is None
        assert Dataset.transform_update_frequency(23) is None
        assert Dataset.transform_update_frequency("15") is None
        assert Dataset.transform_update_frequency("Quarterly") == "90"

    def test_get_set_expected_update_frequency(self, configuration):
        datasetdata = copy.deepcopy(dataset_data)
        dataset = Dataset(datasetdata)
        assert dataset["data_update_frequency"] == "7"
        assert dataset.get_expected_update_frequency() == "Every week"
        dataset.set_expected_update_frequency("every two weeks")
        assert dataset["data_update_frequency"] == "14"
        dataset.set_expected_update_frequency(30)
        assert dataset["data_update_frequency"] == "30"
        dataset.set_expected_update_frequency("Fortnightly")
        assert dataset["data_update_frequency"] == "14"
        assert dataset.get_expected_update_frequency() == "Every two weeks"
        dataset.set_expected_update_frequency("EVERY SIX MONTHS")
        assert dataset["data_update_frequency"] == "180"
        assert dataset.get_expected_update_frequency() == "Every six months"
        dataset.set_expected_update_frequency("90")
        assert dataset["data_update_frequency"] == "90"
        assert dataset.get_expected_update_frequency() == "Every three months"
        with pytest.raises(HDXError):
            dataset.set_expected_update_frequency("lalala")
        with pytest.raises(HDXError):
            dataset.set_expected_update_frequency(9)
        del dataset["data_update_frequency"]
        assert dataset.get_expected_update_frequency() is None

    def test_get_add_tags(self, configuration, vocabulary_read):
        datasetdata = copy.deepcopy(dataset_data)
        dataset = Dataset(datasetdata)
        assert dataset["tags"] == resulttags
        assert dataset.get_tags() == ["conflict", "political violence"]
        dataset.add_tag("LALA")
        assert dataset["tags"] == resulttags
        assert dataset.get_tags() == ["conflict", "political violence"]
        dataset.add_tag("conflict")
        expected = copy.deepcopy(resulttags)
        expected.append(
            {
                "name": "violence and conflict",
                "vocabulary_id": "4381925f-0ae9-44a3-b30d-cae35598757b",
            }
        )
        assert dataset["tags"] == expected
        assert dataset.get_tags() == [
            "conflict",
            "political violence",
            "violence and conflict",
        ]
        dataset.add_tags(
            [
                "desempleo",
                "desocupación",
                "desempleo",
                "conflict-related deaths",
            ]
        )
        assert dataset.get_tags() == [
            "conflict",
            "political violence",
            "violence and conflict",
            "unemployment",
            "fatalities - deaths",
        ]
        dataset.remove_tag("violence and conflict")
        assert dataset.get_tags() == [
            "conflict",
            "political violence",
            "unemployment",
            "fatalities - deaths",
        ]
        del dataset["tags"]
        assert dataset.get_tags() == []
        dataset.add_tag("conflict-related deaths")
        assert dataset["tags"] == [
            {
                "name": "violence and conflict",
                "vocabulary_id": "4381925f-0ae9-44a3-b30d-cae35598757b",
            },
            {
                "name": "fatalities - deaths",
                "vocabulary_id": "4381925f-0ae9-44a3-b30d-cae35598757b",
            },
        ]
        assert dataset.get_tags() == [
            "violence and conflict",
            "fatalities - deaths",
        ]
        dataset.add_tag("conflict-related deaths")
        assert dataset.get_tags() == [
            "violence and conflict",
            "fatalities - deaths",
        ]
        dataset.add_tag("cholera")
        assert dataset.get_tags() == [
            "violence and conflict",
            "fatalities - deaths",
            "cholera",
        ]
        dataset.remove_tag("violence and conflict")
        assert dataset.get_tags() == ["fatalities - deaths", "cholera"]
        dataset.add_tag("cholera")
        assert dataset.get_tags() == ["fatalities - deaths", "cholera"]

    def test_add_clean_tags(self, configuration, vocabulary_read):
        Vocabulary.set_tagsdict(None)
        Vocabulary.read_tags_mappings(failchained=False)
        datasetdata = copy.deepcopy(dataset_data)
        dataset = Dataset(datasetdata)
        assert dataset.get_tags() == ["conflict", "political violence"]
        assert dataset.clean_tags() == (
            ["violence and conflict"],
            ["political violence"],
        )
        dataset.add_tags(["nodeid123", "transportation"])
        assert dataset.get_tags() == [
            "violence and conflict",
            "transportation",
        ]
        dataset["tags"].append(
            {
                "name": "nodeid123",
                "vocabulary_id": "4381925f-0ae9-44a3-b30d-cae35598757b",
            }
        )
        assert dataset.clean_tags() == (
            ["violence and conflict", "transportation"],
            ["nodeid123"],
        )
        assert dataset.get_tags() == [
            "violence and conflict",
            "transportation",
        ]
        dataset.add_tags(["geodata", "points"])
        assert dataset.clean_tags() == (
            ["violence and conflict", "transportation", "geodata"],
            [],
        )
        dataset.add_tag("financial")
        assert dataset.get_tags() == [
            "violence and conflict",
            "transportation",
            "geodata",
        ]
        dataset["tags"].append(
            {
                "name": "financial",
                "vocabulary_id": "4381925f-0ae9-44a3-b30d-cae35598757b",
            }
        )
        assert dataset.clean_tags() == (
            ["violence and conflict", "transportation", "geodata"],
            ["financial"],
        )
        dataset.add_tag("addresses")
        assert dataset.clean_tags() == (
            [
                "violence and conflict",
                "transportation",
                "geodata",
                "3-word addresses",
            ],
            [],
        )
        dataset.remove_tag("3-word addresses")
        assert dataset.get_tags() == [
            "violence and conflict",
            "transportation",
            "geodata",
        ]
        dataset.add_tag("cultivos coca")
        assert dataset.clean_tags() == (
            [
                "violence and conflict",
                "transportation",
                "geodata",
                "food production",
            ],
            [],
        )
        dataset.remove_tag("food production")
        dataset.add_tag("atentados")
        assert dataset.get_tags() == [
            "violence and conflict",
            "transportation",
            "geodata",
            "security incidents",
        ]
        dataset["tags"].append(
            {
                "name": "atentados",
                "vocabulary_id": "4381925f-0ae9-44a3-b30d-cae35598757b",
            }
        )
        assert dataset.clean_tags() == (
            [
                "violence and conflict",
                "transportation",
                "geodata",
                "security incidents",
            ],
            [],
        )
        dataset.remove_tag("security incidents")
        dataset.add_tag("windspeeds")
        assert dataset.clean_tags() == (
            [
                "violence and conflict",
                "transportation",
                "geodata",
                "wind speed",
            ],
            [],
        )
        dataset.add_tag("conservancies")
        assert dataset.get_tags() == [
            "violence and conflict",
            "transportation",
            "geodata",
            "wind speed",
            "protected areas",
        ]
        dataset.remove_tag("transportation")
        dataset.remove_tag("protected areas")
        assert dataset.get_tags() == [
            "violence and conflict",
            "geodata",
            "wind speed",
        ]

    def test_maintainer(self, configuration, user_read):
        dataset = Dataset(dataset_data)
        dataset.set_maintainer("9f3e9973-7dbe-4c65-8820-f48578e3ffea")
        maintainer = dataset.get_maintainer()
        assert maintainer["name"] == "MyUser1"
        user = User(user_data)
        dataset.set_maintainer(user)
        maintainer = dataset.get_maintainer()
        assert maintainer["name"] == "MyUser1"
        with pytest.raises(HDXError):
            dataset.set_maintainer("jpsmith")
        with pytest.raises(HDXError):
            dataset.set_maintainer(123)

    def test_organization(self, configuration, organization_read):
        dataset = Dataset(dataset_data)
        dataset.set_organization("b67e6c74-c185-4f43-b561-0e114a736f19")
        organization = dataset.get_organization()
        assert organization["name"] == "acled"
        organization = Organization(organization_data)
        organization["name"] = "TEST1"
        dataset.set_organization(organization)
        organization = dataset.get_organization()
        assert organization["name"] == "acled"
        with pytest.raises(HDXError):
            dataset.set_organization("123")
        with pytest.raises(HDXError):
            dataset.set_organization(123)

    def test_add_update_delete_showcase(self, configuration, showcase_read):
        datasetdata = copy.deepcopy(dataset_data)
        dataset = Dataset(datasetdata)
        dataset["id"] = "dataset123"
        showcases = dataset.get_showcases()
        assert len(showcases) == 1
        TestDatasetNoncore.association = None
        showcases[0]["id"] = "05e392bf-04e0-4ca6-848c-4e87bba10746"
        dataset.remove_showcase(showcases[0])
        assert TestDatasetNoncore.association == "delete"
        TestDatasetNoncore.association = None
        assert (
            dataset.add_showcase("15e392bf-04e0-4ca6-848c-4e87bba10745")
            is True
        )
        assert TestDatasetNoncore.association == "create"
        TestDatasetNoncore.association = None
        dataset.add_showcases([{"id": "15e392bf-04e0-4ca6-848c-4e87bba10745"}])
        assert TestDatasetNoncore.association == "create"
        TestDatasetNoncore.association = None
        assert (
            dataset.add_showcases(
                [
                    {"id": "15e392bf-04e0-4ca6-848c-4e87bba10745"},
                    {"id": "05e392bf-04e0-4ca6-848c-4e87bba10746"},
                ]
            )
            is False
        )
        assert TestDatasetNoncore.association == "create"
        TestDatasetNoncore.association = None
        assert dataset.add_showcase({"name": "TEST1"}) is True
        assert TestDatasetNoncore.association == "create"
        TestDatasetNoncore.association = None
        with pytest.raises(HDXError):
            dataset.add_showcase("123")
        with pytest.raises(HDXError):
            dataset.add_showcase(123)

    def test_set_quickchart_resource(self, configuration):
        datasetdata = copy.deepcopy(dataset_data)
        resourcesdata = copy.deepcopy(resources_data)
        datasetdata["resources"] = resourcesdata
        dataset = Dataset(datasetdata)
        assert "dataset_preview" not in dataset
        assert (
            dataset.set_quickchart_resource(
                "3d777226-96aa-4239-860a-703389d16d1f"
            )["id"]
            == "3d777226-96aa-4239-860a-703389d16d1f"
        )
        assert dataset["dataset_preview"] == "resource_id"
        resources = dataset.get_resources()
        assert resources[0]["dataset_preview_enabled"] == "False"
        assert resources[1]["dataset_preview_enabled"] == "True"
        assert (
            dataset.set_quickchart_resource(resources[0])["id"]
            == "de6549d8-268b-4dfe-adaf-a4ae5c8510d5"
        )
        assert resources[0]["dataset_preview_enabled"] == "True"
        assert resources[1]["dataset_preview_enabled"] == "False"
        assert (
            dataset.set_quickchart_resource(resources[1].data)["id"]
            == "3d777226-96aa-4239-860a-703389d16d1f"
        )
        assert resources[0]["dataset_preview_enabled"] == "False"
        assert resources[1]["dataset_preview_enabled"] == "True"
        assert (
            dataset.set_quickchart_resource(0)["id"]
            == "de6549d8-268b-4dfe-adaf-a4ae5c8510d5"
        )
        assert resources[0]["dataset_preview_enabled"] == "True"
        assert resources[1]["dataset_preview_enabled"] == "False"
        assert dataset.set_quickchart_resource("12345") is None
        with pytest.raises(HDXError):
            dataset.set_quickchart_resource(True)
        dataset.preview_off()
        assert dataset["dataset_preview"] == "no_preview"
        assert resources[0]["dataset_preview_enabled"] == "False"
        assert resources[1]["dataset_preview_enabled"] == "False"
        assert (
            dataset.set_quickchart_resource("Resource2")["id"]
            == "3d777226-96aa-4239-860a-703389d16d1f"
        )
        assert dataset["dataset_preview"] == "resource_id"
        assert resources[0]["dataset_preview_enabled"] == "False"
        assert resources[1]["dataset_preview_enabled"] == "True"
        assert (
            dataset.set_quickchart_resource({"name": "Resource1"})["id"]
            == "de6549d8-268b-4dfe-adaf-a4ae5c8510d5"
        )
        assert dataset["dataset_preview"] == "resource_id"
        assert resources[0]["dataset_preview_enabled"] == "True"
        assert resources[1]["dataset_preview_enabled"] == "False"

    def test_quickcharts_resource_last(self):
        datasetdata = copy.deepcopy(dataset_data)
        resourcesdata = copy.deepcopy(resources_data)
        datasetdata["resources"] = resourcesdata
        dataset = Dataset(datasetdata)
        assert dataset.quickcharts_resource_last() is False
        resource = {"name": "QuickCharts-resource"}
        dataset.resources.insert(1, resource)
        assert dataset.quickcharts_resource_last() is True
        assert dataset.resources[3]["name"] == resource["name"]
        assert dataset.quickcharts_resource_last() is True

    def test_generate_resource_view(
        self, configuration, vocabulary_update, static_resource_view_yaml
    ):
        datasetdata = copy.deepcopy(dataset_data)
        resourcesdata = copy.deepcopy(resources_data)
        datasetdata["resources"] = resourcesdata
        dataset = Dataset(datasetdata)
        assert "dataset_preview" not in dataset
        resourceview = dataset.generate_resource_view(
            path=static_resource_view_yaml
        )
        hxl_preview_config = json.loads(resourceview["hxl_preview_config"])
        assert resourceview["id"] == "c06b5a0d-1d41-4a74-a196-41c251c76023"
        assert hxl_preview_config["bites"][0]["title"] == "Sum of fatalities"
        assert (
            hxl_preview_config["bites"][1]["title"]
            == "Sum of fatalities grouped by admin1"
        )
        assert (
            hxl_preview_config["bites"][2]["title"]
            == "Sum of fatalities grouped by admin2"
        )
        resourceview = dataset.generate_resource_view(
            path=static_resource_view_yaml, bites_disabled=[False, True, False]
        )
        hxl_preview_config = json.loads(resourceview["hxl_preview_config"])
        assert resourceview["id"] == "c06b5a0d-1d41-4a74-a196-41c251c76023"
        assert hxl_preview_config["bites"][0]["title"] == "Sum of fatalities"
        assert (
            hxl_preview_config["bites"][1]["title"]
            == "Sum of fatalities grouped by admin2"
        )
        resourceview = dataset.generate_resource_view(
            path=static_resource_view_yaml, bites_disabled=[True, True, True]
        )
        assert resourceview is None
        indicators = [
            {
                "code": "1",
                "title": "My1",
                "unit": "ones",
                "description": "This is my one!",
            },
            {
                "code": "2",
                "title": "My2",
                "unit": "twos",
                "aggregate_col": "Agg2",
            },
            {
                "code": "3",
                "title": "My3",
                "description": "This is my three!",
                "date_col": "dt3",
            },
        ]
        resourceview = dataset.generate_resource_view(indicators=indicators)
        hxl_preview_config = json.loads(resourceview["hxl_preview_config"])
        assert resourceview["id"] == "c06b5a0d-1d41-4a74-a196-41c251c76023"
        assert (
            hxl_preview_config["bites"][0]["ingredient"]["filters"][
                "filterWith"
            ][0]["#indicator+code"]
            == "1"
        )
        assert (
            hxl_preview_config["bites"][0]["ingredient"]["description"]
            == "This is my one!"
        )
        assert hxl_preview_config["bites"][0]["uiProperties"]["title"] == "My1"
        assert (
            hxl_preview_config["bites"][0]["computedProperties"]["dataTitle"]
            == "ones"
        )
        assert (
            hxl_preview_config["bites"][1]["ingredient"]["filters"][
                "filterWith"
            ][0]["#indicator+code"]
            == "2"
        )
        assert (
            hxl_preview_config["bites"][1]["ingredient"]["description"] == ""
        )
        assert hxl_preview_config["bites"][1]["uiProperties"]["title"] == "My2"
        assert (
            hxl_preview_config["bites"][1]["computedProperties"]["dataTitle"]
            == "twos"
        )
        assert (
            hxl_preview_config["bites"][1]["ingredient"]["aggregateColumn"]
            == "Agg2"
        )
        assert (
            hxl_preview_config["bites"][2]["ingredient"]["filters"][
                "filterWith"
            ][0]["#indicator+code"]
            == "3"
        )
        assert (
            hxl_preview_config["bites"][2]["ingredient"]["description"]
            == "This is my three!"
        )
        assert (
            hxl_preview_config["bites"][2]["ingredient"]["dateColumn"] == "dt3"
        )
        assert hxl_preview_config["bites"][2]["uiProperties"]["title"] == "My3"
        assert (
            hxl_preview_config["bites"][2]["computedProperties"]["dataTitle"]
            == ""
        )
        resourceview = dataset.generate_resource_view(
            indicators=indicators,
            findreplace={
                "#indicator+code": "#item+code",
                "#indicator+value+num": "#value",
            },
        )
        hxl_preview_config = json.loads(resourceview["hxl_preview_config"])
        assert resourceview["id"] == "c06b5a0d-1d41-4a74-a196-41c251c76023"
        assert (
            hxl_preview_config["bites"][0]["ingredient"]["filters"][
                "filterWith"
            ][0]["#item+code"]
            == "1"
        )
        assert (
            hxl_preview_config["bites"][0]["ingredient"]["valueColumn"]
            == "#value"
        )
        assert dataset.generate_resource_view(indicators=[]) is None
        assert (
            dataset.generate_resource_view(indicators=[None, None, None])
            is None
        )
        assert (
            dataset.generate_resource_view(
                resource="123", path=static_resource_view_yaml
            )
            is None
        )
        del dataset.get_resources()[0]["id"]
        resourceview = dataset.generate_resource_view(
            path=static_resource_view_yaml
        )
        assert "id" not in resourceview
        assert "resource_id" not in resourceview
        assert resourceview["resource_name"] == "Resource1"
        with pytest.raises(IOError):
            dataset.generate_resource_view()

    def test_remove_dates_from_title(self):
        dataset = Dataset()
        with pytest.raises(HDXError):
            dataset.remove_dates_from_title()
        assert "title" not in dataset
        title = "Title with no dates"
        dataset["title"] = title
        assert dataset.remove_dates_from_title() == list()
        assert dataset["title"] == title
        assert "dataset_date" not in dataset
        assert dataset.remove_dates_from_title(set_dataset_date=True) == list()
        title = "ICA Armenia, 2017 - Drought Risk, 1981-2015"
        dataset["title"] = title
        expected = [
            (
                datetime.datetime(1981, 1, 1, 0, 0),
                datetime.datetime(2015, 12, 31, 0, 0),
            ),
            (
                datetime.datetime(2017, 1, 1, 0, 0),
                datetime.datetime(2017, 12, 31, 0, 0),
            ),
        ]
        assert dataset.remove_dates_from_title(change_title=False) == expected
        assert dataset["title"] == title
        assert "dataset_date" not in dataset
        assert dataset.remove_dates_from_title() == expected
        newtitle = "ICA Armenia - Drought Risk"
        assert dataset["title"] == newtitle
        assert "dataset_date" not in dataset
        dataset["title"] = title
        assert (
            dataset.remove_dates_from_title(set_dataset_date=True) == expected
        )
        assert dataset["title"] == newtitle
        assert (
            dataset["dataset_date"]
            == "[1981-01-01T00:00:00 TO 2015-12-31T00:00:00]"
        )
        assert dataset.remove_dates_from_title() == list()
        dataset["title"] = "Mon_State_Village_Tract_Boundaries 9999 2001"
        expected = [
            (
                datetime.datetime(2001, 1, 1, 0, 0),
                datetime.datetime(2001, 12, 31, 0, 0),
            )
        ]
        assert (
            dataset.remove_dates_from_title(set_dataset_date=True) == expected
        )
        assert dataset["title"] == "Mon_State_Village_Tract_Boundaries 9999"
        assert (
            dataset["dataset_date"]
            == "[2001-01-01T00:00:00 TO 2001-12-31T00:00:00]"
        )
        dataset["title"] = "Mon_State_Village_Tract_Boundaries 2001 99"
        assert (
            dataset.remove_dates_from_title(set_dataset_date=True) == expected
        )
        assert dataset["title"] == "Mon_State_Village_Tract_Boundaries 99"
        assert (
            dataset["dataset_date"]
            == "[2001-01-01T00:00:00 TO 2001-12-31T00:00:00]"
        )
        dataset["title"] = "Mon_State_Village_Tract_Boundaries 9999 2001 99"
        assert (
            dataset.remove_dates_from_title(set_dataset_date=True) == expected
        )
        assert dataset["title"] == "Mon_State_Village_Tract_Boundaries 9999 99"
        assert (
            dataset["dataset_date"]
            == "[2001-01-01T00:00:00 TO 2001-12-31T00:00:00]"
        )

    def test_generate_qc_resource_from_rows(self, configuration):
        with temp_dir("test") as folder:
            with Download(user_agent="test") as downloader:
                _, rows = downloader.get_tabular_rows(
                    TestDatasetNoncore.url, dict_form=True, format="csv"
                )
                rows = list(rows)
                dataset = Dataset({"name": "test"})
                qc_filename = "qc_conflict_data_alg.csv"
                resourcedata = {
                    "name": "Conflict Data for Algeria",
                    "description": "Conflict data with HXL tags",
                }
                columnname = "EVENT_ID_CNTY"
                qc_indicator_codes = ["1416RTA", "XXXXRTA", "2231RTA"]
                resource = dataset.generate_qc_resource_from_rows(
                    folder,
                    qc_filename,
                    rows,
                    resourcedata,
                    columnname,
                    TestDatasetNoncore.hxltags,
                    qc_indicator_codes,
                )
                assert resource == {
                    "name": "Conflict Data for Algeria",
                    "description": "Conflict data with HXL tags",
                    "format": "csv",
                    "resource_type": "file.upload",
                    "url_type": "upload",
                }
                assert_files_same(
                    join("tests", "fixtures", "qc_from_rows", qc_filename),
                    join(folder, qc_filename),
                )
                qc_filename = "qc_conflict_data_alg_one_col.csv"
                dataset.generate_qc_resource_from_rows(
                    folder,
                    qc_filename,
                    rows,
                    resourcedata,
                    columnname,
                    TestDatasetNoncore.hxltags,
                    qc_indicator_codes,
                    headers=[columnname],
                )
                assert_files_same(
                    join("tests", "fixtures", "qc_from_rows", qc_filename),
                    join(folder, qc_filename),
                )
                rows = list()
                resource = dataset.generate_qc_resource_from_rows(
                    folder,
                    qc_filename,
                    rows,
                    resourcedata,
                    columnname,
                    TestDatasetNoncore.hxltags,
                    qc_indicator_codes,
                )
                assert resource is None

    def test_download_and_generate_resource(self, configuration):
        with temp_dir("test") as folder:
            filename = "conflict_data_alg.csv"
            resourcedata = {
                "name": "Conflict Data for Algeria",
                "description": "Conflict data with HXL tags",
            }
            admin1s = set()

            def process_row(headers, row):
                row["lala"] = "lala"
                admin1 = row.get("ADMIN1")
                if admin1 is not None:
                    admin1s.add(admin1)
                return row

            dataset = Dataset()
            with Download(user_agent="test") as downloader:
                quickcharts = {
                    "hashtag": "#event+code",
                    "values": ["1416RTA", "XXXXRTA", "2231RTA"],
                    "cutdown": 2,
                }
                success, results = dataset.download_and_generate_resource(
                    downloader,
                    TestDatasetNoncore.url,
                    TestDatasetNoncore.hxltags,
                    folder,
                    filename,
                    resourcedata,
                    header_insertions=[(0, "lala")],
                    row_function=process_row,
                    yearcol="YEAR",
                    quickcharts=quickcharts,
                )
                assert success is True
                assert results == {
                    "startdate": datetime.datetime(2001, 1, 1, 0, 0),
                    "enddate": datetime.datetime(2002, 12, 31, 0, 0),
                    "bites_disabled": [False, True, False],
                    "resource": {
                        "description": "Conflict data with HXL tags",
                        "format": "csv",
                        "name": "Conflict Data for Algeria",
                        "resource_type": "file.upload",
                        "url_type": "upload",
                    },
                    "headers": [
                        "lala",
                        "GWNO",
                        "EVENT_ID_CNTY",
                        "EVENT_ID_NO_CNTY",
                        "EVENT_DATE",
                        "YEAR",
                        "TIME_PRECISION",
                        "EVENT_TYPE",
                        "ACTOR1",
                        "ALLY_ACTOR_1",
                        "INTER1",
                        "ACTOR2",
                        "ALLY_ACTOR_2",
                        "INTER2",
                        "INTERACTION",
                        "COUNTRY",
                        "ADMIN1",
                        "ADMIN2",
                        "ADMIN3",
                        "LOCATION",
                        "LATITUDE",
                        "LONGITUDE",
                        "GEO_PRECISION",
                        "SOURCE",
                        "NOTES",
                        "FATALITIES",
                    ],
                    "rows": [
                        {
                            "lala": "",
                            "GWNO": "",
                            "EVENT_ID_CNTY": "#event+code",
                            "EVENT_ID_NO_CNTY": "",
                            "EVENT_DATE": "#date+occurred",
                            "YEAR": "#date+year",
                            "TIME_PRECISION": "",
                            "EVENT_TYPE": "#event+type",
                            "ACTOR1": "#group+name+first",
                            "ALLY_ACTOR_1": "",
                            "INTER1": "",
                            "ACTOR2": "#group+name+second",
                            "ALLY_ACTOR_2": "",
                            "INTER2": "",
                            "INTERACTION": "",
                            "COUNTRY": "#country+name",
                            "ADMIN1": "#adm1+name",
                            "ADMIN2": "#adm2+name",
                            "ADMIN3": "#adm3+name",
                            "LOCATION": "#loc+name",
                            "LATITUDE": "#geo+lat",
                            "LONGITUDE": "#geo+lon",
                            "GEO_PRECISION": "",
                            "SOURCE": "#meta+source",
                            "NOTES": "#description",
                            "FATALITIES": "#affected+killed",
                        },
                        {
                            "GWNO": "615",
                            "EVENT_ID_CNTY": "1416RTA",
                            "EVENT_ID_NO_CNTY": None,
                            "EVENT_DATE": "18/04/2001",
                            "YEAR": "2001",
                            "TIME_PRECISION": "1",
                            "EVENT_TYPE": "Violence against civilians",
                            "ACTOR1": "Police Forces of Algeria (1999-)",
                            "ALLY_ACTOR_1": None,
                            "INTER1": "1",
                            "ACTOR2": "Civilians (Algeria)",
                            "ALLY_ACTOR_2": "Berber Ethnic Group (Algeria)",
                            "INTER2": "7",
                            "INTERACTION": "17",
                            "COUNTRY": "Algeria",
                            "ADMIN1": "Tizi Ouzou",
                            "ADMIN2": "Beni-Douala",
                            "ADMIN3": None,
                            "LOCATION": "Beni Douala",
                            "LATITUDE": "36.61954",
                            "LONGITUDE": "4.08282",
                            "GEO_PRECISION": "1",
                            "SOURCE": "Associated Press Online",
                            "NOTES": "A Berber student was shot while in police custody at a police station in Beni Douala. He later died on Apr.21.",
                            "FATALITIES": "1",
                            "lala": "lala",
                        },
                        {
                            "GWNO": "615",
                            "EVENT_ID_CNTY": "2229RTA",
                            "EVENT_ID_NO_CNTY": None,
                            "EVENT_DATE": "19/04/2001",
                            "YEAR": "2001",
                            "TIME_PRECISION": "1",
                            "EVENT_TYPE": "Riots/Protests",
                            "ACTOR1": "Rioters (Algeria)",
                            "ALLY_ACTOR_1": "Berber Ethnic Group (Algeria)",
                            "INTER1": "5",
                            "ACTOR2": "Police Forces of Algeria (1999-)",
                            "ALLY_ACTOR_2": None,
                            "INTER2": "1",
                            "INTERACTION": "15",
                            "COUNTRY": "Algeria",
                            "ADMIN1": "Tizi Ouzou",
                            "ADMIN2": "Tizi Ouzou",
                            "ADMIN3": None,
                            "LOCATION": "Tizi Ouzou",
                            "LATITUDE": "36.71183",
                            "LONGITUDE": "4.04591",
                            "GEO_PRECISION": "3",
                            "SOURCE": "Kabylie report",
                            "NOTES": "Riots were reported in numerous villages in Kabylie, resulting in dozens wounded in clashes between protesters and police and significant material damage.",
                            "FATALITIES": "0",
                            "lala": "lala",
                        },
                        {
                            "GWNO": "615",
                            "EVENT_ID_CNTY": "2230RTA",
                            "EVENT_ID_NO_CNTY": None,
                            "EVENT_DATE": "20/04/2001",
                            "YEAR": "2002",
                            "TIME_PRECISION": "1",
                            "EVENT_TYPE": "Riots/Protests",
                            "ACTOR1": "Protesters (Algeria)",
                            "ALLY_ACTOR_1": "Students (Algeria)",
                            "INTER1": "6",
                            "ACTOR2": None,
                            "ALLY_ACTOR_2": None,
                            "INTER2": "0",
                            "INTERACTION": "60",
                            "COUNTRY": "Algeria",
                            "ADMIN1": "Bejaia",
                            "ADMIN2": "Amizour",
                            "ADMIN3": None,
                            "LOCATION": "Amizour",
                            "LATITUDE": "36.64022",
                            "LONGITUDE": "4.90131",
                            "GEO_PRECISION": "1",
                            "SOURCE": "Crisis Group",
                            "NOTES": "Students protested in the Amizour area. At least 3 were later arrested for allegedly insulting gendarmes.",
                            "FATALITIES": None,
                            "lala": "lala",
                        },
                        {
                            "GWNO": "615",
                            "EVENT_ID_CNTY": "2231RTA",
                            "EVENT_ID_NO_CNTY": None,
                            "EVENT_DATE": "21/04/2001",
                            "YEAR": "2001",
                            "TIME_PRECISION": "1",
                            "EVENT_TYPE": "Riots/Protests",
                            "ACTOR1": "Rioters (Algeria)",
                            "ALLY_ACTOR_1": "Berber Ethnic Group (Algeria)",
                            "INTER1": "5",
                            "ACTOR2": "Police Forces of Algeria (1999-)",
                            "ALLY_ACTOR_2": None,
                            "INTER2": "1",
                            "INTERACTION": "15",
                            "COUNTRY": "Algeria",
                            "ADMIN1": "Bejaia",
                            "ADMIN2": "Amizour",
                            "ADMIN3": None,
                            "LOCATION": "Amizour",
                            "LATITUDE": "36.64022",
                            "LONGITUDE": "4.90131",
                            "GEO_PRECISION": "1",
                            "SOURCE": "Kabylie report",
                            "NOTES": "Rioters threw molotov cocktails, rocks and burning tires at gendarmerie stations in Beni Douala, El-Kseur and Amizour.",
                            "FATALITIES": "0",
                            "lala": "lala",
                        },
                    ],
                    "qc_resource": {
                        "description": "Cut down data for QuickCharts",
                        "format": "csv",
                        "name": "QuickCharts-Conflict Data for Algeria",
                        "resource_type": "file.upload",
                        "url_type": "upload",
                    },
                    "qcheaders": [
                        "EVENT_ID_CNTY",
                        "EVENT_DATE",
                        "YEAR",
                        "EVENT_TYPE",
                        "ACTOR1",
                        "ACTOR2",
                        "COUNTRY",
                        "ADMIN1",
                        "ADMIN2",
                        "ADMIN3",
                        "LOCATION",
                        "LATITUDE",
                        "LONGITUDE",
                        "SOURCE",
                        "NOTES",
                        "FATALITIES",
                    ],
                    "qcrows": [
                        {
                            "EVENT_ID_CNTY": "#event+code",
                            "EVENT_DATE": "#date+occurred",
                            "YEAR": "#date+year",
                            "EVENT_TYPE": "#event+type",
                            "ACTOR1": "#group+name+first",
                            "ACTOR2": "#group+name+second",
                            "COUNTRY": "#country+name",
                            "ADMIN1": "#adm1+name",
                            "ADMIN2": "#adm2+name",
                            "ADMIN3": "#adm3+name",
                            "LOCATION": "#loc+name",
                            "LATITUDE": "#geo+lat",
                            "LONGITUDE": "#geo+lon",
                            "SOURCE": "#meta+source",
                            "NOTES": "#description",
                            "FATALITIES": "#affected+killed",
                        },
                        {
                            "EVENT_ID_CNTY": "1416RTA",
                            "EVENT_DATE": "18/04/2001",
                            "YEAR": "2001",
                            "EVENT_TYPE": "Violence against civilians",
                            "ACTOR1": "Police Forces of Algeria (1999-)",
                            "ACTOR2": "Civilians (Algeria)",
                            "COUNTRY": "Algeria",
                            "ADMIN1": "Tizi Ouzou",
                            "ADMIN2": "Beni-Douala",
                            "ADMIN3": None,
                            "LOCATION": "Beni Douala",
                            "LATITUDE": "36.61954",
                            "LONGITUDE": "4.08282",
                            "SOURCE": "Associated Press Online",
                            "NOTES": "A Berber student was shot while in police custody at a police station in Beni Douala. He later died on Apr.21.",
                            "FATALITIES": "1",
                        },
                        {
                            "EVENT_ID_CNTY": "2231RTA",
                            "EVENT_DATE": "21/04/2001",
                            "YEAR": "2001",
                            "EVENT_TYPE": "Riots/Protests",
                            "ACTOR1": "Rioters (Algeria)",
                            "ACTOR2": "Police Forces of Algeria (1999-)",
                            "COUNTRY": "Algeria",
                            "ADMIN1": "Bejaia",
                            "ADMIN2": "Amizour",
                            "ADMIN3": None,
                            "LOCATION": "Amizour",
                            "LATITUDE": "36.64022",
                            "LONGITUDE": "4.90131",
                            "SOURCE": "Kabylie report",
                            "NOTES": "Rioters threw molotov cocktails, rocks and burning tires at gendarmerie stations in Beni Douala, El-Kseur and Amizour.",
                            "FATALITIES": "0",
                        },
                    ],
                }
                assert (
                    dataset["dataset_date"]
                    == "[2001-01-01T00:00:00 TO 2002-12-31T00:00:00]"
                )
                assert admin1s == {"Bejaia", "Tizi Ouzou"}
                resources = dataset.get_resources()
                assert resources == [
                    {
                        "name": "Conflict Data for Algeria",
                        "description": "Conflict data with HXL tags",
                        "format": "csv",
                        "resource_type": "file.upload",
                        "url_type": "upload",
                    },
                    {
                        "name": "QuickCharts-Conflict Data for Algeria",
                        "description": "Cut down data for QuickCharts",
                        "format": "csv",
                        "resource_type": "file.upload",
                        "url_type": "upload",
                    },
                ]
                assert_files_same(
                    join("tests", "fixtures", "gen_resource", filename),
                    join(folder, filename),
                )
                qc_filename = f"qc_{filename}"
                assert_files_same(
                    join("tests", "fixtures", "gen_resource", qc_filename),
                    join(folder, qc_filename),
                )

                success, results = dataset.download_and_generate_resource(
                    downloader,
                    TestDatasetNoncore.url,
                    TestDatasetNoncore.hxltags,
                    folder,
                    filename,
                    resourcedata,
                    header_insertions=[(0, "lala")],
                    row_function=process_row,
                    datecol="EVENT_DATE",
                    quickcharts=quickcharts,
                )
                assert success is True
                assert (
                    dataset["dataset_date"]
                    == "[2001-04-18T00:00:00 TO 2001-04-21T00:00:00]"
                )

                quickcharts = {
                    "hashtag": "#event+code",
                    "values": ["1416RTA", "2230RTA", "2231RTA"],
                    "numeric_hashtag": "#affected+killed",
                    "cutdown": 2,
                    "cutdownhashtags": ["#event+code"],
                }
                success, results = dataset.download_and_generate_resource(
                    downloader,
                    TestDatasetNoncore.url,
                    TestDatasetNoncore.hxltags,
                    folder,
                    filename,
                    resourcedata,
                    header_insertions=[(0, "lala")],
                    row_function=process_row,
                    yearcol="YEAR",
                    quickcharts=quickcharts,
                )
                assert success is True
                assert results == {
                    "startdate": datetime.datetime(2001, 1, 1, 0, 0),
                    "enddate": datetime.datetime(2002, 12, 31, 0, 0),
                    "bites_disabled": [False, True, False],
                    "resource": {
                        "description": "Conflict data with HXL tags",
                        "format": "csv",
                        "name": "Conflict Data for Algeria",
                        "resource_type": "file.upload",
                        "url_type": "upload",
                    },
                    "headers": [
                        "lala",
                        "GWNO",
                        "EVENT_ID_CNTY",
                        "EVENT_ID_NO_CNTY",
                        "EVENT_DATE",
                        "YEAR",
                        "TIME_PRECISION",
                        "EVENT_TYPE",
                        "ACTOR1",
                        "ALLY_ACTOR_1",
                        "INTER1",
                        "ACTOR2",
                        "ALLY_ACTOR_2",
                        "INTER2",
                        "INTERACTION",
                        "COUNTRY",
                        "ADMIN1",
                        "ADMIN2",
                        "ADMIN3",
                        "LOCATION",
                        "LATITUDE",
                        "LONGITUDE",
                        "GEO_PRECISION",
                        "SOURCE",
                        "NOTES",
                        "FATALITIES",
                    ],
                    "rows": [
                        {
                            "lala": "",
                            "GWNO": "",
                            "EVENT_ID_CNTY": "#event+code",
                            "EVENT_ID_NO_CNTY": "",
                            "EVENT_DATE": "#date+occurred",
                            "YEAR": "#date+year",
                            "TIME_PRECISION": "",
                            "EVENT_TYPE": "#event+type",
                            "ACTOR1": "#group+name+first",
                            "ALLY_ACTOR_1": "",
                            "INTER1": "",
                            "ACTOR2": "#group+name+second",
                            "ALLY_ACTOR_2": "",
                            "INTER2": "",
                            "INTERACTION": "",
                            "COUNTRY": "#country+name",
                            "ADMIN1": "#adm1+name",
                            "ADMIN2": "#adm2+name",
                            "ADMIN3": "#adm3+name",
                            "LOCATION": "#loc+name",
                            "LATITUDE": "#geo+lat",
                            "LONGITUDE": "#geo+lon",
                            "GEO_PRECISION": "",
                            "SOURCE": "#meta+source",
                            "NOTES": "#description",
                            "FATALITIES": "#affected+killed",
                        },
                        {
                            "GWNO": "615",
                            "EVENT_ID_CNTY": "1416RTA",
                            "EVENT_ID_NO_CNTY": None,
                            "EVENT_DATE": "18/04/2001",
                            "YEAR": "2001",
                            "TIME_PRECISION": "1",
                            "EVENT_TYPE": "Violence against civilians",
                            "ACTOR1": "Police Forces of Algeria (1999-)",
                            "ALLY_ACTOR_1": None,
                            "INTER1": "1",
                            "ACTOR2": "Civilians (Algeria)",
                            "ALLY_ACTOR_2": "Berber Ethnic Group (Algeria)",
                            "INTER2": "7",
                            "INTERACTION": "17",
                            "COUNTRY": "Algeria",
                            "ADMIN1": "Tizi Ouzou",
                            "ADMIN2": "Beni-Douala",
                            "ADMIN3": None,
                            "LOCATION": "Beni Douala",
                            "LATITUDE": "36.61954",
                            "LONGITUDE": "4.08282",
                            "GEO_PRECISION": "1",
                            "SOURCE": "Associated Press Online",
                            "NOTES": "A Berber student was shot while in police custody at a police station in Beni Douala. He later died on Apr.21.",
                            "FATALITIES": "1",
                            "lala": "lala",
                        },
                        {
                            "GWNO": "615",
                            "EVENT_ID_CNTY": "2229RTA",
                            "EVENT_ID_NO_CNTY": None,
                            "EVENT_DATE": "19/04/2001",
                            "YEAR": "2001",
                            "TIME_PRECISION": "1",
                            "EVENT_TYPE": "Riots/Protests",
                            "ACTOR1": "Rioters (Algeria)",
                            "ALLY_ACTOR_1": "Berber Ethnic Group (Algeria)",
                            "INTER1": "5",
                            "ACTOR2": "Police Forces of Algeria (1999-)",
                            "ALLY_ACTOR_2": None,
                            "INTER2": "1",
                            "INTERACTION": "15",
                            "COUNTRY": "Algeria",
                            "ADMIN1": "Tizi Ouzou",
                            "ADMIN2": "Tizi Ouzou",
                            "ADMIN3": None,
                            "LOCATION": "Tizi Ouzou",
                            "LATITUDE": "36.71183",
                            "LONGITUDE": "4.04591",
                            "GEO_PRECISION": "3",
                            "SOURCE": "Kabylie report",
                            "NOTES": "Riots were reported in numerous villages in Kabylie, resulting in dozens wounded in clashes between protesters and police and significant material damage.",
                            "FATALITIES": "0",
                            "lala": "lala",
                        },
                        {
                            "GWNO": "615",
                            "EVENT_ID_CNTY": "2230RTA",
                            "EVENT_ID_NO_CNTY": None,
                            "EVENT_DATE": "20/04/2001",
                            "YEAR": "2002",
                            "TIME_PRECISION": "1",
                            "EVENT_TYPE": "Riots/Protests",
                            "ACTOR1": "Protesters (Algeria)",
                            "ALLY_ACTOR_1": "Students (Algeria)",
                            "INTER1": "6",
                            "ACTOR2": None,
                            "ALLY_ACTOR_2": None,
                            "INTER2": "0",
                            "INTERACTION": "60",
                            "COUNTRY": "Algeria",
                            "ADMIN1": "Bejaia",
                            "ADMIN2": "Amizour",
                            "ADMIN3": None,
                            "LOCATION": "Amizour",
                            "LATITUDE": "36.64022",
                            "LONGITUDE": "4.90131",
                            "GEO_PRECISION": "1",
                            "SOURCE": "Crisis Group",
                            "NOTES": "Students protested in the Amizour area. At least 3 were later arrested for allegedly insulting gendarmes.",
                            "FATALITIES": None,
                            "lala": "lala",
                        },
                        {
                            "GWNO": "615",
                            "EVENT_ID_CNTY": "2231RTA",
                            "EVENT_ID_NO_CNTY": None,
                            "EVENT_DATE": "21/04/2001",
                            "YEAR": "2001",
                            "TIME_PRECISION": "1",
                            "EVENT_TYPE": "Riots/Protests",
                            "ACTOR1": "Rioters (Algeria)",
                            "ALLY_ACTOR_1": "Berber Ethnic Group (Algeria)",
                            "INTER1": "5",
                            "ACTOR2": "Police Forces of Algeria (1999-)",
                            "ALLY_ACTOR_2": None,
                            "INTER2": "1",
                            "INTERACTION": "15",
                            "COUNTRY": "Algeria",
                            "ADMIN1": "Bejaia",
                            "ADMIN2": "Amizour",
                            "ADMIN3": None,
                            "LOCATION": "Amizour",
                            "LATITUDE": "36.64022",
                            "LONGITUDE": "4.90131",
                            "GEO_PRECISION": "1",
                            "SOURCE": "Kabylie report",
                            "NOTES": "Rioters threw molotov cocktails, rocks and burning tires at gendarmerie stations in Beni Douala, El-Kseur and Amizour.",
                            "FATALITIES": "0",
                            "lala": "lala",
                        },
                    ],
                    "qc_resource": {
                        "description": "Cut down data for QuickCharts",
                        "format": "csv",
                        "name": "QuickCharts-Conflict Data for Algeria",
                        "resource_type": "file.upload",
                        "url_type": "upload",
                    },
                    "qcheaders": ["EVENT_ID_CNTY", "FATALITIES"],
                    "qcrows": [
                        {
                            "EVENT_ID_CNTY": "#event+code",
                            "FATALITIES": "#affected+killed",
                        },
                        {"EVENT_ID_CNTY": "1416RTA", "FATALITIES": "1"},
                        {"EVENT_ID_CNTY": "2231RTA", "FATALITIES": "0"},
                    ],
                }

                def process_year(row):
                    year = row["YEAR"]
                    if year == "2002":
                        return None
                    startdate, enddate = parse_date_range(year)
                    return {"startdate": startdate, "enddate": enddate}

                del quickcharts["hashtag"]
                del quickcharts["numeric_hashtag"]
                success, results = dataset.download_and_generate_resource(
                    downloader,
                    TestDatasetNoncore.url,
                    TestDatasetNoncore.hxltags,
                    folder,
                    filename,
                    resourcedata,
                    header_insertions=[(0, "lala")],
                    row_function=process_row,
                    date_function=process_year,
                    quickcharts=quickcharts,
                )
                assert success is True
                assert results["startdate"] == datetime.datetime(
                    2001, 1, 1, 0, 0
                )
                assert results["enddate"] == datetime.datetime(
                    2001, 12, 31, 0, 0
                )
                assert (
                    dataset["dataset_date"]
                    == "[2001-01-01T00:00:00 TO 2001-12-31T00:00:00]"
                )
                assert_files_same(
                    join(
                        "tests",
                        "fixtures",
                        "gen_resource",
                        f"min_{qc_filename}",
                    ),
                    join(folder, qc_filename),
                )

                with pytest.raises(HDXError):
                    dataset.download_and_generate_resource(
                        downloader,
                        TestDatasetNoncore.url,
                        TestDatasetNoncore.hxltags,
                        folder,
                        filename,
                        resourcedata,
                        yearcol="YEAR",
                        date_function=process_year,
                    )
                success, results = dataset.download_and_generate_resource(
                    downloader,
                    TestDatasetNoncore.url,
                    TestDatasetNoncore.hxltags,
                    folder,
                    filename,
                    resourcedata,
                    header_insertions=[(0, "lala")],
                    row_function=process_row,
                )
                assert success is True
                url = "https://raw.githubusercontent.com/OCHA-DAP/hdx-python-api/master/tests/fixtures/empty.csv"
                success, results = dataset.download_and_generate_resource(
                    downloader,
                    url,
                    TestDatasetNoncore.hxltags,
                    folder,
                    filename,
                    resourcedata,
                    header_insertions=[(0, "lala")],
                    row_function=process_row,
                    yearcol="YEAR",
                )
                assert success is False
                url = "https://raw.githubusercontent.com/OCHA-DAP/hdx-python-api/master/tests/fixtures/gen_resource/test_data_no_data.csv"
                success, results = dataset.download_and_generate_resource(
                    downloader,
                    url,
                    TestDatasetNoncore.hxltags,
                    folder,
                    filename,
                    resourcedata,
                    header_insertions=[(0, "lala")],
                    row_function=process_row,
                    quickcharts=quickcharts,
                )
                assert success is False
                url = "https://raw.githubusercontent.com/OCHA-DAP/hdx-python-api/master/tests/fixtures/gen_resource/test_data_no_years.csv"
                success, results = dataset.download_and_generate_resource(
                    downloader,
                    url,
                    TestDatasetNoncore.hxltags,
                    folder,
                    filename,
                    resourcedata,
                    header_insertions=[(0, "lala")],
                    row_function=process_row,
                    yearcol="YEAR",
                )
                assert success is False

    def test_load_save_to_json(self, vocabulary_read):
        with temp_dir(
            "LoadSaveDatasetJSON",
            delete_on_success=True,
            delete_on_failure=False,
        ) as temp_folder:
            name = "mydatasetname"
            dataset = Dataset(
                {"name": name, "title": "title", "notes": "description"}
            )
            maintainer = "196196be-6037-4488-8b71-d786adf4c081"
            dataset.set_maintainer(maintainer)
            dataset.set_organization("fb7c2910-6080-4b66-8b4f-0be9b6dc4d8e")
            start_date = "2020-02-09"
            end_date = "2020-10-20"
            dataset.set_date_of_dataset(start_date, end_date)
            expected_update_frequency = "Every day"
            dataset.set_expected_update_frequency(expected_update_frequency)
            dataset.set_subnational(False)
            tags = ["hxl", "aid funding"]
            dataset.add_tags(tags)
            resource_name = "filename.csv"
            resourcedata = {
                "name": resource_name,
                "description": "resource description",
                "format": "csv",
                "url": "http://lala",
            }
            dataset.add_update_resource(resourcedata)
            path = join(temp_folder, "dataset.json")
            dataset.save_to_json(path)
            dataset = Dataset.load_from_json(path)
            assert dataset["name"] == name
            assert dataset["maintainer"] == maintainer
            dateinfo = dataset.get_date_of_dataset()
            assert dateinfo["startdate_str"][:10] == start_date
            assert dateinfo["enddate_str"][:10] == end_date
            assert (
                dataset.get_expected_update_frequency()
                == expected_update_frequency
            )
            assert dataset.get_tags() == tags
            assert dataset.get_resource()["name"] == resource_name

            save_text("null", path)
            dataset = Dataset.load_from_json(path)
            assert dataset is None



"""Locations in HDX"""
from typing import Dict, List, Optional, Tuple

from hdx.api.configuration import Configuration


class Locations:
    """Methods to help with countries and continents"""

    _validlocations = None

    @classmethod
    def validlocations(cls, configuration=None) -> List[Dict]:
        """
        Read valid locations from HDX

        Args:
            configuration (Optional[Configuration]): HDX configuration. Defaults to global configuration.

        Returns:
            List[Dict]: A list of valid locations
        """
        if cls._validlocations is None:
            if configuration is None:
                configuration = Configuration.read()
            cls._validlocations = configuration.call_remoteckan(
                "group_list", {"all_fields": True}
            )
        return cls._validlocations

    @classmethod
    def set_validlocations(cls, locations: List[Dict]) -> None:
        """
        Set valid locations using list of dictionaries of form {'name': 'zmb', 'title', 'Zambia'}

        Args:
            locations (List[Dict]): List of dictionaries of form {'name': 'zmb', 'title', 'Zambia'}

        Returns:
            None
        """
        cls._validlocations = locations

    @classmethod
    def get_location_from_HDX_code(
        cls,
        code: str,
        locations: Optional[List[Dict]] = None,
        configuration: Optional[Configuration] = None,
    ) -> Optional[str]:
        """Get location from HDX location code

        Args:
            code (str): code for which to get location name
            locations (Optional[List[Dict]]): Valid locations list. Defaults to list downloaded from HDX.
            configuration (Optional[Configuration]): HDX configuration. Defaults to global configuration.

        Returns:
            Optional[str]: location name
        """
        if locations is None:
            locations = cls.validlocations(configuration)
        for locdict in locations:
            if code.upper() == locdict["name"].upper():
                return locdict["title"]
        return None

    @classmethod
    def get_HDX_code_from_location(
        cls,
        location: str,
        locations: Optional[List[Dict]] = None,
        configuration: Optional[Configuration] = None,
    ) -> Optional[str]:
        """Get HDX code for location

        Args:
            location (str): Location for which to get HDX code
            locations (Optional[List[Dict]]): Valid locations list. Defaults to list downloaded from HDX.
            configuration (Optional[Configuration]): HDX configuration. Defaults to global configuration.

        Returns:
            Optional[str]: HDX code or None
        """
        if locations is None:
            locations = cls.validlocations(configuration)
        locationupper = location.upper()
        for locdict in locations:
            locationcode = locdict["name"].upper()
            if locationupper == locationcode:
                return locationcode

        for locdict in locations:
            if locationupper == locdict["title"].upper():
                return locdict["name"].upper()
        return None

    @classmethod
    def get_HDX_code_from_location_partial(
        cls,
        location: str,
        locations: Optional[List[Dict]] = None,
        configuration: Optional[Configuration] = None,
    ) -> Tuple[Optional[str], bool]:
        """Get HDX code for location

        Args:
            location (str): Location for which to get HDX code
            locations (Optional[List[Dict]]): Valid locations list. Defaults to list downloaded from HDX.
            configuration (Optional[Configuration]): HDX configuration. Defaults to global configuration.

        Returns:
            Tuple[Optional[str], bool]: HDX code and if the match is exact or (None, False) for no match
        """
        hdx_code = cls.get_HDX_code_from_location(
            location, locations, configuration
        )

        if hdx_code is not None:
            return hdx_code, True

        if locations is None:
            locations = cls.validlocations(configuration)
        locationupper = location.upper()
        for locdict in locations:
            locationname = locdict["title"].upper()
            if locationupper in locationname or locationname in locationupper:
                return locdict["name"].upper(), False

        return None, False



import copy
import json


class MockResponse:
    def __init__(self, status_code, text):
        self.status_code = status_code
        self.text = text

    def json(self):
        return json.loads(self.text)


resultgroups = [
    {
        "description": "",
        "name": "dza",
        "image_display_url": "",
        "display_name": "Algeria",
        "id": "dza",
        "title": "Algeria",
    },
    {
        "description": "",
        "name": "zwe",
        "image_display_url": "",
        "display_name": "Zimbabwe",
        "id": "zwe",
        "title": "Zimbabwe",
    },
]

resulttags = [
    {
        "state": "active",
        "display_name": "conflict",
        "vocabulary_id": "4381925f-0ae9-44a3-b30d-cae35598757b",
        "id": "1dae41e5-eacd-4fa5-91df-8d80cf579e53",
        "name": "conflict",
    },
    {
        "state": "active",
        "display_name": "political violence",
        "vocabulary_id": "4381925f-0ae9-44a3-b30d-cae35598757b",
        "id": "aaafc63b-2234-48e3-8ccc-198d7cf0f3f3",
        "name": "political violence",
    },
]

dataset_data = {
    "name": "MyDataset1",
    "title": "MyDataset1",
    "dataset_date": "06/04/2016",  # has to be MM/DD/YYYY
    "groups": resultgroups,
    "owner_org": "My Org",
    "author": "AN Other",
    "author_email": "another@another.com",
    "maintainer": "AN Other",
    "maintainer_email": "another@another.com",
    "license_id": "cc-by-sa",
    "subnational": "1",
    "notes": "some notes",
    "caveats": "some caveats",
    "data_update_frequency": "7",
    "methodology": "other",
    "methodology_other": "methodology description",
    "dataset_source": "Mr Org",
    "package_creator": "AN Other",
    "private": False,
    "url": None,
    "state": "active",
    "tags": resulttags,
}

resources_data = [
    {
        "id": "de6549d8-268b-4dfe-adaf-a4ae5c8510d5",
        "description": "Resource1",
        "name": "Resource1",
        "url": "http://resource1.xlsx",
        "format": "xlsx",
    },
    {
        "id": "3d777226-96aa-4239-860a-703389d16d1f",
        "description": "Resource2",
        "name": "Resource2",
        "url": "http://resource2.csv",
        "format": "csv",
    },
    {
        "id": "3d777226-96aa-4239-860a-703389d16d1g",
        "description": "Resource2",
        "name": "Resource2",
        "url": "http://resource2.xls",
        "format": "xls",
    },
]

organization_data = {
    "name": "MyOrganization1",
    "title": "Humanitarian Organization",
    "description": "We do humanitarian work",
}

user_data = {
    "name": "MyUser1",
    "email": "xxx@yyy.com",
    "password": "xxx",
    "fullname": "xxx xxx",
    "about": "Data Scientist",
    "capacity": "admin",
}

dataset_resultdict = {
    "resources": [
        {
            "revision_id": "43765383-1fce-471f-8166-d6c8660cc8a9",
            "cache_url": None,
            "datastore_active": False,
            "format": "xlsx",
            "webstore_url": None,
            "last_modified": None,
            "tracking_summary": {"recent": 0, "total": 0},
            "id": "de6549d8-268b-4dfe-adaf-a4ae5c8510d5",
            "webstore_last_updated": None,
            "mimetype": None,
            "state": "active",
            "created": "2016-06-07T08:57:27.367939",
            "description": "Resource1",
            "position": 0,
            "hash": "",
            "package_id": "6f36a41c-f126-4b18-aaaf-6c2ddfbc5d4d",
            "name": "Resource1",
            "url": "http://resource1.xlsx",
            "resource_type": "api",
            "url_type": "api",
            "size": None,
            "mimetype_inner": None,
            "cache_last_updated": None,
        },
        {
            "revision_id": "387e5d1a-50ca-4039-b5a7-f7b6b88d0f2b",
            "cache_url": None,
            "datastore_active": False,
            "format": "csv",
            "webstore_url": None,
            "last_modified": None,
            "tracking_summary": {"recent": 0, "total": 0},
            "id": "3d777226-96aa-4239-860a-703389d16d1f",
            "webstore_last_updated": None,
            "mimetype": None,
            "state": "active",
            "created": "2016-06-07T08:57:27.367959",
            "description": "Resource2",
            "position": 1,
            "hash": "",
            "package_id": "6f36a41c-f126-4b18-aaaf-6c2ddfbc5d4d",
            "name": "Resource2",
            "url": "http://resource2_csv.zip",
            "resource_type": "api",
            "url_type": "api",
            "size": None,
            "mimetype_inner": None,
            "cache_last_updated": None,
        },
        {
            "revision_id": "387e5d1a-50ca-4039-b5a7-f7b6b88d0f2c",
            "cache_url": None,
            "datastore_active": False,
            "format": "geojson",
            "webstore_url": None,
            "last_modified": None,
            "tracking_summary": {"recent": 0, "total": 0},
            "id": "3d777226-96aa-4239-860a-703389d16d1g",
            "webstore_last_updated": None,
            "mimetype": None,
            "state": "active",
            "created": "2016-06-07T08:57:27.367959",
            "description": "Resource2",
            "position": 2,
            "hash": "",
            "package_id": "6f36a41c-f126-4b18-aaaf-6c2ddfbc5d4d",
            "name": "Resource2",
            "url": "http://resource2_csv.zip",
            "resource_type": "api",
            "url_type": "api",
            "size": None,
            "mimetype_inner": None,
            "cache_last_updated": None,
        },
    ],
    "isopen": True,
    "caveats": "Various",
    "revision_id": "032833ca-c403-40cc-8b86-69d5a6eecb1b",
    "url": None,
    "author": "acled",
    "metadata_created": "2016-03-23T14:28:48.664205",
    "license_url": "http://www.opendefinition.org/licenses/cc-by-sa",
    "relationships_as_object": [],
    "creator_user_id": "154de241-38d6-47d3-a77f-0a9848a61df3",
    "methodology_other": "This page contains information.",
    "subnational": "1",
    "maintainer_email": "me@me.com",
    "license_title": "Creative Commons Attribution Share-Alike",
    "title": "MyDataset",
    "private": False,
    "maintainer": "8b84230c-e04a-43ec-99e5-41307a203a2f",
    "methodology": "Other",
    "num_tags": 4,
    "license_id": "cc-by-sa",
    "tracking_summary": {"recent": 32, "total": 178},
    "relationships_as_subject": [],
    "owner_org": "b67e6c74-c185-4f43-b561-0e114a736f19",
    "id": "6f36a41c-f126-4b18-aaaf-6c2ddfbc5d4d",
    "dataset_source": "ACLED",
    "type": "dataset",
    "notes": "Notes",
    "organization": {
        "revision_id": "684f3eee-b708-4f91-bd22-7860d4eca423",
        "description": "MyOrganisation",
        "name": "acled",
        "type": "organization",
        "image_url": "",
        "approval_status": "approved",
        "state": "active",
        "title": "MyOrganisation",
        "created": "2015-01-09T14:44:54.006612",
        "id": "b67e6c74-c185-4f43-b561-0e114a736f19",
        "is_organization": True,
    },
    "state": "active",
    "author_email": "me@me.com",
    "package_creator": "someone",
    "num_resources": 2,
    "total_res_downloads": 4,
    "name": "MyDataset1",
    "metadata_modified": "2016-06-09T12:49:33.854367",
    "groups": resultgroups,
    "data_update_frequency": "7",
    "tags": resulttags,
    "version": None,
    "solr_additions": '{"countries": ["Algeria", "Zimbabwe"]}',
    "dataset_date": "06/04/2016",
}


def dataset_mockshow(url, datadict):
    if "show" not in url:
        return MockResponse(
            404,
            '{"success": false, "error": {"message": "TEST ERROR: Not show", "__type": "TEST ERROR: Not Show Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=package_show"}',
        )
    if "resource_show" in url:
        result = json.dumps(resources_data[0])
        return MockResponse(
            200,
            '{"success": true, "result": %s, "help": "http://test-data.humdata.org/api/3/action/help_show?name=resource_show"}'
            % result,
        )
    else:
        if datadict["id"] == "TEST1":
            result = json.dumps(dataset_resultdict)
            return MockResponse(
                200,
                '{"success": true, "result": %s, "help": "http://test-data.humdata.org/api/3/action/help_show?name=package_show"}'
                % result,
            )
        if datadict["id"] == "DatasetExist":
            result = json.dumps(dataset_resultdict)
            return MockResponse(
                200,
                '{"success": true, "result": %s, "help": "http://test-data.humdata.org/api/3/action/help_show?name=package_show"}'
                % result,
            )
        if datadict["id"] == "TEST2":
            return MockResponse(
                404,
                '{"success": false, "error": {"message": "Not found", "__type": "Not Found Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=package_show"}',
            )
        if datadict["id"] == "TEST3":
            return MockResponse(
                200,
                '{"success": false, "error": {"message": "Not found", "__type": "Not Found Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=package_show"}',
            )
        if datadict["id"] == "TEST4":
            resultdictcopy = copy.deepcopy(dataset_resultdict)
            resultdictcopy["id"] = "TEST4"
            result = json.dumps(resultdictcopy)
            return MockResponse(
                200,
                '{"success": true, "result": %s, "help": "http://test-data.humdata.org/api/3/action/help_show?name=package_show"}'
                % result,
            )

    return MockResponse(
        404,
        '{"success": false, "error": {"message": "Not found", "__type": "Not Found Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=package_show"}',
    )



"""User Tests"""
import copy
import json
from os.path import join

import pytest
from hdx.utilities.dictandlist import merge_two_dictionaries
from hdx.utilities.loader import load_yaml

from hdx.api.configuration import Configuration
from hdx.data.hdxobject import HDXError
from hdx.data.user import User

from . import MockResponse, user_data

resultdict = {
    "openid": None,
    "about": "Data Scientist",
    "apikey": "31e86726-2993-4d82-be93-3d2133c81d94",
    "display_name": "xxx",
    "name": "MyUser1",
    "created": "2017-03-22T09:55:55.871573",
    "email_hash": "9e2fc15d8cc65f43136f9dcef05fa184",
    "email": "xxx@yyy.com",
    "sysadmin": False,
    "activity_streams_email_notifications": False,
    "state": "active",
    "number_of_edits": 0,
    "fullname": "xxx xxx",
    "id": "9f3e9973-7dbe-4c65-8820-f48578e3ffea",
    "number_created_packages": 0,
}
orgdict = load_yaml(join("tests", "fixtures", "organization_show_results.yml"))
orgdict2 = copy.deepcopy(orgdict)
del orgdict2["users"]
del orgdict2["packages"]
del orgdict2["extras"]
orglist = [orgdict2]

result2dict = copy.deepcopy(resultdict)
result2dict["email"] = "aaa@bbb.com"
user_list = [resultdict, result2dict]

user_autocomplete = [
    {
        "fullname": "Fake Test",
        "id": "2ea425bf-130d-4b63-b755-5cc5435b4b75",
        "name": "another-fake-tester",
    }
]


def user_mockshow(url, datadict):
    if "show" not in url:
        return MockResponse(
            404,
            '{"success": false, "error": {"message": "TEST ERROR: Not show", "__type": "TEST ERROR: Not Show Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=user_show"}',
        )
    result = json.dumps(resultdict)
    if (
        datadict["id"] == "9f3e9973-7dbe-4c65-8820-f48578e3ffea"
        or datadict["id"] == "MyUser1"
    ):
        return MockResponse(
            200,
            '{"success": true, "result": %s, "help": "http://test-data.humdata.org/api/3/action/help_show?name=user_show"}'
            % result,
        )
    if datadict["id"] == "TEST2":
        return MockResponse(
            404,
            '{"success": false, "error": {"message": "Not found", "__type": "Not Found Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=user_show"}',
        )
    if datadict["id"] == "TEST3":
        return MockResponse(
            200,
            '{"success": false, "error": {"message": "Not found", "__type": "Not Found Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=user_show"}',
        )
    return MockResponse(
        404,
        '{"success": false, "error": {"message": "Not found", "__type": "Not Found Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=user_show"}',
    )


def mocklist(url):
    if "list" not in url:
        return MockResponse(
            404,
            '{"success": false, "error": {"message": "TEST ERROR: Not all", "__type": "TEST ERROR: Not All Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=user_list"}',
        )
    return MockResponse(
        200,
        '{"success": true, "result": %s, "help": "http://test-data.humdata.org/api/3/action/help_show?name=user_list"}'
        % json.dumps(user_list),
    )


class TestUser:
    smtp_initargs = {
        "host": "localhost",
        "port": 123,
        "local_hostname": "mycomputer.fqdn.com",
        "timeout": 3,
        "source_address": ("machine", 456),
    }
    username = "user"
    password = "pass"
    email_config_dict = {
        "connection_type": "ssl",
        "username": username,
        "password": password,
    }
    subject = "hello"
    text_body = "hello there"
    html_body = """\
<html>
  <head></head>
  <body>
    <p>Hi!<br>
    </p>
  </body>
</html>
"""
    sender = "me@gmail.com"
    mail_options = ["a", "b"]
    rcpt_options = [1, 2]
    email_config_dict.update(smtp_initargs)

    @pytest.fixture(scope="class")
    def static_yaml(self):
        return join("tests", "fixtures", "config", "hdx_user_static.yml")

    @pytest.fixture(scope="class")
    def static_json(self):
        return join("tests", "fixtures", "config", "hdx_user_static.json")

    @pytest.fixture(scope="function")
    def read(self):
        class MockSession:
            @staticmethod
            def post(url, data, headers, files, allow_redirects, auth=None):
                datadict = json.loads(data.decode("utf-8"))
                return user_mockshow(url, datadict)

        Configuration.read().remoteckan().session = MockSession()

    @pytest.fixture(scope="function")
    def post_create(self):
        class MockSession:
            @staticmethod
            def post(url, data, headers, files, allow_redirects, auth=None):
                datadict = json.loads(data.decode("utf-8"))
                if "show" in url:
                    return user_mockshow(url, datadict)
                if "create" not in url:
                    return MockResponse(
                        404,
                        '{"success": false, "error": {"message": "TEST ERROR: Not create", "__type": "TEST ERROR: Not Create Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=user_create"}',
                    )

                resultdictcopy = copy.deepcopy(resultdict)
                resultdictcopy["state"] = datadict["state"]
                result = json.dumps(resultdictcopy)
                if datadict["name"] == "MyUser1":
                    return MockResponse(
                        200,
                        '{"success": true, "result": %s, "help": "http://test-data.humdata.org/api/3/action/help_show?name=user_create"}'
                        % result,
                    )
                if datadict["name"] == "MyUser2":
                    return MockResponse(
                        404,
                        '{"success": false, "error": {"message": "Not found", "__type": "Not Found Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=user_create"}',
                    )
                if datadict["name"] == "MyUser3":
                    return MockResponse(
                        200,
                        '{"success": false, "error": {"message": "Not found", "__type": "Not Found Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=user_create"}',
                    )

                return MockResponse(
                    404,
                    '{"success": false, "error": {"message": "Not found", "__type": "Not Found Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=user_create"}',
                )

        Configuration.read().remoteckan().session = MockSession()

    @pytest.fixture(scope="function")
    def post_update(self):
        class MockSession:
            @staticmethod
            def post(url, data, headers, files, allow_redirects, auth=None):
                datadict = json.loads(data.decode("utf-8"))
                if "show" in url:
                    return user_mockshow(url, datadict)
                if "update" not in url:
                    return MockResponse(
                        404,
                        '{"success": false, "error": {"message": "TEST ERROR: Not update", "__type": "TEST ERROR: Not Update Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=user_update"}',
                    )
                resultdictcopy = copy.deepcopy(resultdict)
                merge_two_dictionaries(resultdictcopy, datadict)

                result = json.dumps(resultdictcopy)
                if datadict["name"] == "MyUser1":
                    return MockResponse(
                        200,
                        '{"success": true, "result": %s, "help": "http://test-data.humdata.org/api/3/action/help_show?name=user_update"}'
                        % result,
                    )
                if datadict["name"] == "MyUser2":
                    return MockResponse(
                        404,
                        '{"success": false, "error": {"message": "Not found", "__type": "Not Found Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=user_update"}',
                    )
                if datadict["name"] == "MyUser3":
                    return MockResponse(
                        200,
                        '{"success": false, "error": {"message": "Not found", "__type": "Not Found Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=user_update"}',
                    )

                return MockResponse(
                    404,
                    '{"success": false, "error": {"message": "Not found", "__type": "Not Found Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=user_update"}',
                )

        Configuration.read().remoteckan().session = MockSession()

    @pytest.fixture(scope="function")
    def post_delete(self):
        class MockSession:
            @staticmethod
            def post(url, data, headers, files, allow_redirects, auth=None):
                decodedata = data.decode("utf-8")
                datadict = json.loads(decodedata)
                if "show" in url:
                    return user_mockshow(url, datadict)
                if "delete" not in url:
                    return MockResponse(
                        404,
                        '{"success": false, "error": {"message": "TEST ERROR: Not delete", "__type": "TEST ERROR: Not Delete Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=user_delete"}',
                    )
                if datadict["id"] == "9f3e9973-7dbe-4c65-8820-f48578e3ffea":
                    return MockResponse(
                        200,
                        '{"success": true, "result": %s, "help": "http://test-data.humdata.org/api/3/action/help_show?name=user_delete"}'
                        % decodedata,
                    )

                return MockResponse(
                    404,
                    '{"success": false, "error": {"message": "Not found", "__type": "Not Found Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=user_delete"}',
                )

        Configuration.read().remoteckan().session = MockSession()

    @pytest.fixture(scope="function")
    def post_list(self):
        class MockSession:
            @staticmethod
            def post(url, data, headers, files, allow_redirects, auth=None):
                json.loads(data.decode("utf-8"))
                return mocklist(url)

        Configuration.read().remoteckan().session = MockSession()

    @pytest.fixture(scope="function")
    def post_listorgs(self):
        class MockSession:
            @staticmethod
            def post(url, data, headers, files, allow_redirects, auth=None):
                decodedata = data.decode("utf-8")
                datadict = json.loads(decodedata)
                if "user" in url:
                    if "show" in url:
                        return user_mockshow(url, datadict)
                    elif "list" in url:
                        return MockResponse(
                            200,
                            '{"success": true, "result": %s, "help": "http://test-data.humdata.org/api/3/action/help_show?name=organization_list"}'
                            % json.dumps(orglist),
                        )
                elif "organization" in url:
                    if "show" in url:
                        result = json.dumps(orgdict)
                        if (
                            datadict["id"]
                            == "b67e6c74-c185-4f43-b561-0e114a736f19"
                            or datadict["id"] == "TEST1"
                        ):
                            return MockResponse(
                                200,
                                '{"success": true, "result": %s, "help": "http://test-data.humdata.org/api/3/action/help_show?name=user_show"}'
                                % result,
                            )

        Configuration.read().remoteckan().session = MockSession()

    @pytest.fixture(scope="function")
    def post_autocomplete(self):
        class MockSession:
            @staticmethod
            def post(url, data, headers, files, allow_redirects, auth=None):
                decodedata = data.decode("utf-8")
                datadict = json.loads(decodedata)
                if "autocomplete" not in url or "fake" not in datadict["q"]:
                    return MockResponse(
                        404,
                        '{"success": false, "error": {"message": "TEST ERROR: Not autocomplete", "__type": "TEST ERROR: Not Autocomplete Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=user_autocomplete"}',
                    )
                result = json.dumps(user_autocomplete)
                return MockResponse(
                    200,
                    '{"success": true, "result": %s, "help": "http://test-data.humdata.org/api/3/action/help_show?name=user_autocomplete"}'
                    % result,
                )

        Configuration.read().remoteckan().session = MockSession()

    def test_read_from_hdx(self, configuration, read, mocksmtp):
        user = User.read_from_hdx("9f3e9973-7dbe-4c65-8820-f48578e3ffea")
        assert user["id"] == "9f3e9973-7dbe-4c65-8820-f48578e3ffea"
        assert user["name"] == "MyUser1"
        user = User.read_from_hdx("TEST2")
        assert user is None
        user = User.read_from_hdx("TEST3")
        assert user is None
        config = Configuration.read()
        config.setup_emailer(email_config_dict=TestUser.email_config_dict)
        user = User.read_from_hdx("9f3e9973-7dbe-4c65-8820-f48578e3ffea")
        user.email(
            TestUser.subject,
            TestUser.text_body,
            html_body=TestUser.html_body,
            sender=TestUser.sender,
            mail_options=TestUser.mail_options,
            rcpt_options=TestUser.rcpt_options,
        )
        email = config.emailer()
        assert email.server.type == "smtpssl"
        assert email.server.initargs == TestUser.smtp_initargs
        assert email.server.username == TestUser.username
        assert email.server.password == TestUser.password
        assert email.server.sender == TestUser.sender
        assert email.server.recipients == ["xxx@yyy.com"]
        assert "Content-Type: multipart/alternative;" in email.server.msg
        assert (
            """\
MIME-Version: 1.0
Subject: hello
From: me@gmail.com
To: xxx@yyy.com"""
            in email.server.msg
        )
        assert (
            """\
Content-Type: text/plain; charset="us-ascii"
MIME-Version: 1.0
Content-Transfer-Encoding: 7bit

hello there"""
            in email.server.msg
        )
        assert (
            """\
Content-Type: text/html; charset="us-ascii"
MIME-Version: 1.0
Content-Transfer-Encoding: 7bit

<html>
  <head></head>
  <body>
    <p>Hi!<br>
    </p>
  </body>
</html>"""
            in email.server.msg
        )
        assert email.server.send_args == {
            "mail_options": ["a", "b"],
            "rcpt_options": [1, 2],
        }

    def test_create_in_hdx(self, configuration, post_create):
        user = User()
        with pytest.raises(HDXError):
            user.create_in_hdx()
        user["id"] = "9f3e9973-7dbe-4c65-8820-f48578e3ffea"
        user["name"] = "LALA"
        with pytest.raises(HDXError):
            user.create_in_hdx()

        data = copy.deepcopy(user_data)
        user = User(data)
        user.create_in_hdx()
        assert user["id"] == "9f3e9973-7dbe-4c65-8820-f48578e3ffea"
        assert user["capacity"] == "admin"
        assert user["state"] == "active"

        data["name"] = "MyUser2"
        user = User(data)
        with pytest.raises(HDXError):
            user.create_in_hdx()

        data["name"] = "MyUser3"
        user = User(data)
        with pytest.raises(HDXError):
            user.create_in_hdx()

    def test_update_in_hdx(self, configuration, post_update):
        user = User()
        user["id"] = "NOTEXIST"
        with pytest.raises(HDXError):
            user.update_in_hdx()
        user["name"] = "LALA"
        with pytest.raises(HDXError):
            user.update_in_hdx()

        user = User.read_from_hdx("9f3e9973-7dbe-4c65-8820-f48578e3ffea")
        assert user["id"] == "9f3e9973-7dbe-4c65-8820-f48578e3ffea"
        assert user["about"] == "Data Scientist"

        user["about"] = "IMO"
        user["id"] = "9f3e9973-7dbe-4c65-8820-f48578e3ffea"
        user["name"] = "MyUser1"
        user["capacity"] = "member"
        user.update_in_hdx()
        assert user["id"] == "9f3e9973-7dbe-4c65-8820-f48578e3ffea"
        assert user["about"] == "IMO"
        assert user["capacity"] == "member"
        assert user["state"] == "active"

        user["id"] = "NOTEXIST"
        with pytest.raises(HDXError):
            user.update_in_hdx()

        del user["id"]
        with pytest.raises(HDXError):
            user.update_in_hdx()

        data = copy.deepcopy(user_data)
        data["name"] = "MyUser1"
        data["id"] = "9f3e9973-7dbe-4c65-8820-f48578e3ffea"
        user = User(data)
        user.create_in_hdx()
        assert user["id"] == "9f3e9973-7dbe-4c65-8820-f48578e3ffea"
        assert user["about"] == "Data Scientist"
        assert user["state"] == "active"

    def test_delete_from_hdx(self, configuration, post_delete):
        user = User.read_from_hdx("9f3e9973-7dbe-4c65-8820-f48578e3ffea")
        user.delete_from_hdx()
        del user["id"]
        with pytest.raises(HDXError):
            user.delete_from_hdx()

    def test_update_yaml(self, configuration, static_yaml):
        data = copy.deepcopy(user_data)
        user = User(data)
        assert user["name"] == "MyUser1"
        assert user["about"] == "Data Scientist"
        user.update_from_yaml(static_yaml)
        assert user["name"] == "MyUser1"
        assert user["about"] == "IMO"

    def test_update_json(self, configuration, static_json):
        data = copy.deepcopy(user_data)
        user = User(data)
        assert user["name"] == "MyUser1"
        assert user["about"] == "Data Scientist"
        user.update_from_json(static_json)
        assert user["name"] == "MyUser1"
        assert user["about"] == "other"

    def test_get_all_users(self, configuration, post_list, mocksmtp):
        users = User.get_all_users()
        assert len(users) == 2
        config = Configuration.read()
        config.setup_emailer(email_config_dict=TestUser.email_config_dict)
        User.email_users(
            users,
            TestUser.subject,
            TestUser.text_body,
            html_body=TestUser.html_body,
            sender=TestUser.sender,
            cc=users,
            bcc=users,
            mail_options=TestUser.mail_options,
            rcpt_options=TestUser.rcpt_options,
        )
        email = config.emailer()
        assert email.server.type == "smtpssl"
        assert email.server.initargs == TestUser.smtp_initargs
        assert email.server.username == TestUser.username
        assert email.server.password == TestUser.password
        assert email.server.sender == TestUser.sender
        assert email.server.recipients == [
            "xxx@yyy.com",
            "aaa@bbb.com",
            "xxx@yyy.com",
            "aaa@bbb.com",
            "xxx@yyy.com",
            "aaa@bbb.com",
        ]
        assert "Content-Type: multipart/alternative;" in email.server.msg
        assert (
            """\
MIME-Version: 1.0
Subject: hello
From: me@gmail.com
To: xxx@yyy.com, aaa@bbb.com
Cc: xxx@yyy.com, aaa@bbb.com"""
            in email.server.msg
        )
        assert (
            """\
Content-Type: text/plain; charset="us-ascii"
MIME-Version: 1.0
Content-Transfer-Encoding: 7bit

hello there"""
            in email.server.msg
        )
        assert (
            """\
Content-Type: text/html; charset="us-ascii"
MIME-Version: 1.0
Content-Transfer-Encoding: 7bit

<html>
  <head></head>
  <body>
    <p>Hi!<br>
    </p>
  </body>
</html>"""
            in email.server.msg
        )
        assert email.server.send_args == {
            "mail_options": ["a", "b"],
            "rcpt_options": [1, 2],
        }
        with pytest.raises(ValueError):
            User.email_users(
                list(),
                TestUser.subject,
                TestUser.text_body,
                sender=TestUser.sender,
                mail_options=TestUser.mail_options,
                rcpt_options=TestUser.rcpt_options,
            )

    def test_get_organizations(self, configuration, post_listorgs):
        user = User.read_from_hdx("9f3e9973-7dbe-4c65-8820-f48578e3ffea")
        organizations = user.get_organizations()
        assert organizations[0]["id"] == "b67e6c74-c185-4f43-b561-0e114a736f19"

    def test_autocomplete(self, configuration, post_autocomplete):
        assert User.autocomplete("fake%20test") == user_autocomplete
